# Copyright (c) 2015 The Johns Hopkins University/Applied Physics Laboratory
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Exception for keymgr
"""

from tacker._i18n import _
from tacker.common.exceptions import TackerException


class Forbidden(TackerException):
    message = _("You are not authorized to complete this action.")


class KeyManagerError(TackerException):
    message = _("Key manager error: %(reason)s")


class ManagedObjectNotFoundError(TackerException):
    message = _("Key not found, uuid: %(uuid)s")


class AuthTypeInvalidError(TackerException):
    message = _("Invalid auth_type was specified, auth_type: %(type)s")


class InsufficientCredentialDataError(TackerException):
    message = _('Insufficient credential data was provided, either '
                '"token" must be set in the passed conf, or a context '
                'with an "auth_token" property must be passed.')



# Copyright 2016 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import sqlalchemy
from sqlalchemy import and_
from sqlalchemy import create_engine, pool
from sqlalchemy import inspect

from oslo_utils import timeutils

from tacker._i18n import _
from tacker.common import exceptions


GRANULARITY = {'days': 86400, 'hours': 3600, 'minutes': 60, 'seconds': 1}


def _generate_associated_tables_map(inspector):
    assoc_map = {}
    table_names = inspector.get_table_names()
    for t in table_names:
        fk_list = inspector.get_foreign_keys(t)
        for fk in fk_list:
            k = str(fk['referred_table'])
            v = str(fk['constrained_columns'][0])
            if k not in assoc_map:
                assoc_map[k] = {str(t): v}
            else:
                assoc_map[k][str(t)] = v
    assoc_keys = assoc_map.keys()
    for k, v in assoc_map.items():
        for k1 in v:
            if k1 in assoc_keys:
                del assoc_map[k][k1]
    return assoc_map


def _purge_resource_tables(t, meta, engine, time_line, assoc_map):
    table_load = sqlalchemy.Table(t, meta, autoload=True)
    table_del_query = table_load.delete().where(
        table_load.c.deleted_at <= time_line)
    if t in assoc_map:
        select_id_query = sqlalchemy.select([table_load.c.id]).where(
            table_load.c.deleted_at <= time_line)
        resource_ids = [i[0] for i in list(engine.execute(select_id_query))]
        if resource_ids:
            for key, val in assoc_map[t].items():
                assoc_table_load = sqlalchemy.Table(key, meta, autoload=True)
                assoc_table_del_query = assoc_table_load.delete().where(
                    assoc_table_load.c[val].in_(resource_ids))
                engine.execute(assoc_table_del_query)
    engine.execute(table_del_query)


def _purge_events_table(meta, engine, time_line):
    tname = "events"
    event_table_load = sqlalchemy.Table(tname, meta, autoload=True)
    event_select_query = sqlalchemy.select(
        [event_table_load.c.resource_id]
    ).where(
        and_(event_table_load.c.event_type == 'DELETE',
             event_table_load.c.timestamp <= time_line
             )
    )
    resource_ids = [i[0] for i in list(engine.execute(event_select_query))]
    if resource_ids:
        event_delete_query = event_table_load.delete().where(
            event_table_load.c.resource_id.in_(resource_ids)
        )
        engine.execute(event_delete_query)


def purge_deleted(tacker_config, table_name, age, granularity='days'):
    try:
        age = int(age)
    except ValueError:
        msg = _("'%s' - age should be an integer") % age
        raise exceptions.InvalidInput(error_message=msg)
    if age < 0:
        msg = _("'%s' - age should be a positive integer") % age
        raise exceptions.InvalidInput(error_message=msg)

    if granularity not in GRANULARITY:
        msg = _("'%s' granularity should be days, hours, minutes, "
                "or seconds") % granularity
        raise exceptions.InvalidInput(error_message=msg)

    age *= GRANULARITY[granularity]

    time_line = timeutils.utcnow() - datetime.timedelta(seconds=age)
    engine = get_engine(tacker_config)
    meta = sqlalchemy.MetaData()
    meta.bind = engine
    inspector = inspect(engine)
    assoc_map = _generate_associated_tables_map(inspector)

    if table_name == 'events':
        _purge_events_table(meta, engine, time_line)
    elif table_name == 'all':
        _purge_events_table(meta, engine, time_line)
        for t in ['vnf', 'vnfd', 'vims']:
            _purge_resource_tables(t, meta, engine, time_line, assoc_map)
    else:
        _purge_resource_tables(table_name, meta, engine, time_line, assoc_map)


def get_engine(tacker_config):
    return create_engine(tacker_config.database.connection,
                         poolclass=pool.NullPool)



# Copyright (C) 2019 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_policy import policy

from tacker.policies import base


VNFPKGM = 'os_nfv_orchestration_api:vnf_packages:%s'

rules = [
    policy.DocumentedRuleDefault(
        name=VNFPKGM % 'create',
        check_str=base.RULE_ADMIN_OR_OWNER,
        description="Creates a vnf package.",
        operations=[
            {
                'method': 'POST',
                'path': '/vnf_packages'
            }
        ]),
    policy.DocumentedRuleDefault(
        name=VNFPKGM % 'show',
        check_str=base.RULE_ADMIN_OR_OWNER,
        description="Show a vnf package.",
        operations=[
            {
                'method': 'GET',
                'path': '/vnf_packages/{vnf_package_id}'
            }
        ]),
    policy.DocumentedRuleDefault(
        name=VNFPKGM % 'index',
        check_str=base.RULE_ADMIN_OR_OWNER,
        description="List all vnf packages.",
        operations=[
            {
                'method': 'GET',
                'path': '/vnf_packages/'
            }
        ]),
    policy.DocumentedRuleDefault(
        name=VNFPKGM % 'delete',
        check_str=base.RULE_ADMIN_OR_OWNER,
        description="Delete a vnf package.",
        operations=[
            {
                'method': 'DELETE',
                'path': '/vnf_packages/{vnf_package_id}'
            }
        ]),
    policy.DocumentedRuleDefault(
        name=VNFPKGM % 'fetch_package_content',
        check_str=base.RULE_ADMIN_OR_OWNER,
        description="fetch the contents of an on-boarded VNF Package",
        operations=[
            {
                'method': 'GET',
                'path': '/vnf_packages/{vnf_package_id}/'
                        'package_content'
            }
        ]),
    policy.DocumentedRuleDefault(
        name=VNFPKGM % 'upload_package_content',
        check_str=base.RULE_ADMIN_OR_OWNER,
        description="upload a vnf package content.",
        operations=[
            {
                'method': 'PUT',
                'path': '/vnf_packages/{vnf_package_id}/'
                        'package_content'
            }
        ]),
    policy.DocumentedRuleDefault(
        name=VNFPKGM % 'upload_from_uri',
        check_str=base.RULE_ADMIN_OR_OWNER,
        description="upload a vnf package content from uri.",
        operations=[
            {
                'method': 'POST',
                'path': '/vnf_packages/{vnf_package_id}/package_content/'
                        'upload_from_uri'
            }
        ]),
    policy.DocumentedRuleDefault(
        name=VNFPKGM % 'patch',
        check_str=base.RULE_ADMIN_OR_OWNER,
        description="update information of vnf package.",
        operations=[
            {
                'method': 'PATCH',
                'path': '/vnf_packages/{vnf_package_id}'
            }
        ]),
    policy.DocumentedRuleDefault(
        name=VNFPKGM % 'get_vnf_package_vnfd',
        check_str=base.RULE_ADMIN_OR_OWNER,
        description="reads the content of the VNFD within a VNF package.",
        operations=[
            {
                'method': 'GET',
                'path': '/vnf_packages/{vnf_package_id}/vnfd'
            }
        ]),
    policy.DocumentedRuleDefault(
        name=VNFPKGM % 'fetch_artifact',
        check_str=base.RULE_ADMIN_OR_OWNER,
        description="reads the content of the artifact within a VNF package.",
        operations=[
            {
                'method': 'GET',
                'path': '/vnf_packages/{vnfPkgId}/artifacts/{artifactPath}'
            }
        ]),
]


def list_rules():
    return rules



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from tacker.sol_refactored.common import exceptions as sol_ex
from tacker.sol_refactored.common import vnfd_utils


class AbstractUserData(metaclass=abc.ABCMeta):
    """Definition of each method

    Args:
        req: Request dict for each API
             (ex. InstantiateVnfRequest for instantiate)
        inst: VnfInstance dict
        grant_req: GrantRequest dict
        grant: Grant dict
        tmp_csar_dir: directory path that csar contents are extracted

    Returns:
        dict of parameters for create/update heat stack.
        see the example of userdata_default.py.
    """

    @staticmethod
    @abc.abstractmethod
    def instantiate(req, inst, grant_req, grant, tmp_csar_dir):
        raise sol_ex.UserDataClassNotImplemented()

    @staticmethod
    @abc.abstractmethod
    def scale(req, inst, grant_req, grant, tmp_csar_dir):
        raise sol_ex.UserDataClassNotImplemented()

    @staticmethod
    @abc.abstractmethod
    def change_ext_conn(req, inst, grant_req, grant, tmp_csar_dir):
        raise sol_ex.UserDataClassNotImplemented()

    @staticmethod
    @abc.abstractmethod
    def heal(req, inst, grant_req, grant, tmp_csar_dir):
        raise sol_ex.UserDataClassNotImplemented()


def get_vnfd(vnfd_id, csar_dir):
    vnfd = vnfd_utils.Vnfd(vnfd_id)
    vnfd.init_from_csar_dir(csar_dir)
    return vnfd


def init_nfv_dict(hot_template):
    get_params = []

    def _get_get_param(prop):
        if isinstance(prop, dict):
            for key, value in prop.items():
                if key == 'get_param':
                    get_params.append(value)
                else:
                    _get_get_param(value)
        elif isinstance(prop, list):
            for value in prop:
                _get_get_param(value)

    for res in hot_template.get('resources', {}).values():
        _get_get_param(res.get('properties', {}))

    nfv = {}

    for param in get_params:
        if (not isinstance(param, list) or len(param) < 4 or
                param[0] != 'nfv'):
            continue
        parent = nfv
        for item in param[1:-1]:
            parent.setdefault(item, {})
            parent = parent[item]
        parent[param[-1]] = None

    # TODO(oda-g): enhance to handle list
    # NOTE: List is not considered here and only 'fixed_ips' is treated as
    # list in userdata_default.py at the moment.
    # Note that if handling list is enhanced, userdata_default.py is
    # necessary to modify.
    return nfv


def get_param_flavor(vdu_name, flavour_id, vnfd, grant):
    # try to get from grant
    if 'vimAssets' in grant:
        assets = grant['vimAssets']
        if 'computeResourceFlavours' in assets:
            flavours = assets['computeResourceFlavours']
            for flavour in flavours:
                if flavour['vnfdVirtualComputeDescId'] == vdu_name:
                    return flavour['vimFlavourId']

    # if specified in VNFD, use it
    # NOTE: if not found. parameter is set to None.
    #       may be error when stack create
    return vnfd.get_compute_flavor(flavour_id, vdu_name)


def get_param_image(vdu_name, flavour_id, vnfd, grant):
    # try to get from grant
    if 'vimAssets' in grant:
        assets = grant['vimAssets']
        if 'softwareImages' in assets:
            images = assets['softwareImages']
            for image in images:
                if image['vnfdSoftwareImageId'] == vdu_name:
                    return image['vimSoftwareImageId']

    # if specified in VNFD, use it
    # NOTE: if not found. parameter is set to None.
    #       may be error when stack create
    sw_images = vnfd.get_sw_image(flavour_id)
    for name, image in sw_images.items():
        if name == vdu_name:
            return image


def get_param_zone(vdu_name, grant_req, grant):
    if 'zones' not in grant or 'addResources' not in grant:
        return

    for res in grant['addResources']:
        if 'zoneId' not in res:
            continue
        for req_res in grant_req['addResources']:
            if req_res['id'] == res['resourceDefinitionId']:
                if req_res.get('resourceTemplateId') == vdu_name:
                    for zone in grant['zones']:
                        if zone['id'] == res['zoneId']:  # must be found
                            return zone['zoneId']


def get_current_capacity(vdu_name, inst):
    count = 0
    inst_vnfcs = (inst.get('instantiatedVnfInfo', {})
                      .get('vnfcResourceInfo', []))
    for inst_vnfc in inst_vnfcs:
        if inst_vnfc['vduId'] == vdu_name:
            count += 1

    return count


def get_param_capacity(vdu_name, inst, grant_req):
    # NOTE: refer grant_req here since interpretation of VNFD was done when
    # making grant_req.
    count = get_current_capacity(vdu_name, inst)

    add_reses = grant_req.get('addResources', [])
    for res_def in add_reses:
        if (res_def['type'] == 'COMPUTE' and
                res_def['resourceTemplateId'] == vdu_name):
            count += 1

    rm_reses = grant_req.get('removeResources', [])
    for res_def in rm_reses:
        if (res_def['type'] == 'COMPUTE' and
                res_def['resourceTemplateId'] == vdu_name):
            count -= 1

    return count


def _get_fixed_ips_from_extcp(extcp):
    fixed_ips = []
    for cp_conf in extcp['cpConfig'].values():
        if 'cpProtocolData' not in cp_conf:
            continue
        for prot_data in cp_conf['cpProtocolData']:
            if 'ipOverEthernet' not in prot_data:
                continue
            if 'ipAddresses' not in prot_data['ipOverEthernet']:
                continue
            for ip in prot_data['ipOverEthernet']['ipAddresses']:
                data = {}
                if 'fixedAddresses' in ip:
                    # pick up only one ip address
                    data['ip_address'] = str(ip['fixedAddresses'][0])
                if 'subnetId' in ip:
                    data['subnet'] = ip['subnetId']
                if data:
                    fixed_ips.append(data)
    return fixed_ips


def get_param_network(cp_name, grant, req):
    # see grant first then instantiateVnfRequest
    vls = grant.get('extVirtualLinks', []) + req.get('extVirtualLinks', [])
    for vl in vls:
        for extcp in vl['extCps']:
            if extcp['cpdId'] == cp_name:
                return vl['resourceId']


def get_param_fixed_ips(cp_name, grant, req):
    # see grant first then instantiateVnfRequest
    vls = grant.get('extVirtualLinks', []) + req.get('extVirtualLinks', [])
    for vl in vls:
        for extcp in vl['extCps']:
            if extcp['cpdId'] == cp_name:
                return _get_fixed_ips_from_extcp(extcp)


def get_param_network_from_inst(cp_name, inst):
    for vl in inst['instantiatedVnfInfo'].get('extVirtualLinkInfo', []):
        for extcp in vl.get('currentVnfExtCpData', []):
            if extcp['cpdId'] == cp_name:
                return vl['resourceHandle']['resourceId']


def get_param_fixed_ips_from_inst(cp_name, inst):
    for vl in inst['instantiatedVnfInfo'].get('extVirtualLinkInfo', []):
        for extcp in vl.get('currentVnfExtCpData', []):
            if extcp['cpdId'] == cp_name:
                return _get_fixed_ips_from_extcp(extcp)


def apply_ext_managed_vls(hot_dict, req, grant):
    # see grant first then instantiateVnfRequest
    mgd_vls = (grant.get('extManagedVirtualLinks', []) +
               req.get('extManagedVirtualLinks', []))

    # NOTE: refer HOT only here, not refer VNFD.
    # HOT and VNFD must be consistent.

    for mgd_vl in mgd_vls:
        vl_name = mgd_vl['vnfVirtualLinkDescId']
        network_id = mgd_vl['resourceId']
        get_res = {'get_resource': vl_name}

        def _change(item):
            if not isinstance(item, dict):
                return
            for key, value in item.items():
                if value == get_res:
                    item[key] = network_id
                else:
                    _change(value)

        del_reses = []
        for res_name, res_data in hot_dict.get('resources', {}).items():
            # delete network definition
            if res_name == vl_name:
                del_reses.append(res_name)

            # delete subnet definition
            if res_data['type'] == 'OS::Neutron::Subnet':
                net = (res_data.get('properties', {})
                               .get('network', {})
                               .get('get_resource'))
                if net == vl_name:
                    del_reses.append(res_name)

            # change '{get_resource: vl_name}' to network_id
            _change(res_data)

        for res_name in del_reses:
            hot_dict['resources'].pop(res_name)



# Copyright 2016 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Add error_reason to device

Revision ID: acf941e54075
Revises: 5246a6bd410f
Create Date: 2016-04-07 23:53:56.623647

"""

# flake8: noqa: E402

# revision identifiers, used by Alembic.
revision = 'acf941e54075'
down_revision = '5246a6bd410f'

from alembic import op
import sqlalchemy as sa


def upgrade(active_plugins=None, options=None):
    op.add_column('devices', sa.Column('error_reason',
        sa.Text(), nullable=True))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from datetime import datetime

from oslo_db.exception import DBDuplicateEntry
from oslo_utils import strutils
from oslo_utils import timeutils
from oslo_utils import uuidutils
from sqlalchemy.orm import exc as orm_exc
from sqlalchemy import sql

from tacker.common import exceptions
from tacker.db.common_services import common_services_db_plugin
from tacker.db import db_base
from tacker.db.nfvo import nfvo_db
from tacker.db.vnfm import vnfm_db
from tacker.extensions import nfvo
from tacker import manager
from tacker.plugins.common import constants


VIM_ATTRIBUTES = ('id', 'type', 'tenant_id', 'name', 'description',
                  'placement_attr', 'shared', 'is_default',
                  'created_at', 'updated_at', 'status', 'extra')

VIM_AUTH_ATTRIBUTES = ('auth_url', 'vim_project', 'password', 'auth_cred')


class NfvoPluginDb(nfvo.NFVOPluginBase, db_base.CommonDbMixin):

    def __init__(self):
        super(NfvoPluginDb, self).__init__()
        self._cos_db_plg = common_services_db_plugin.CommonServicesPluginDb()

    @property
    def _core_plugin(self):
        return manager.TackerManager.get_plugin()

    def _make_vim_dict(self, vim_db, fields=None, mask_password=True):
        res = dict((key, vim_db[key]) for key in VIM_ATTRIBUTES)
        vim_auth_db = vim_db.vim_auth
        res['auth_url'] = vim_auth_db[0].auth_url
        res['vim_project'] = vim_auth_db[0].vim_project
        res['auth_cred'] = vim_auth_db[0].auth_cred
        res['auth_cred']['password'] = vim_auth_db[0].password
        if mask_password:
            res['auth_cred'] = strutils.mask_dict_password(res['auth_cred'])
        return self._fields(res, fields)

    def _fields(self, resource, fields):
        if fields:
            return dict(((key, item) for key, item in resource.items()
                         if key in fields))
        return resource

    def _get_resource(self, context, model, id):
        try:
            return self._get_by_id(context, model, id)
        except orm_exc.NoResultFound:
            if issubclass(model, nfvo_db.Vim):
                raise nfvo.VimNotFoundException(vim_id=id)
            else:
                raise

    def create_vim(self, context, vim):
        self._validate_default_vim(context, vim)
        vim_cred = vim['auth_cred']

        try:
            with context.session.begin(subtransactions=True):
                vim_db = nfvo_db.Vim(
                    id=vim.get('id'),
                    type=vim.get('type'),
                    tenant_id=vim.get('tenant_id'),
                    name=vim.get('name'),
                    description=vim.get('description'),
                    placement_attr=vim.get('placement_attr'),
                    is_default=vim.get('is_default'),
                    status=vim.get('status'),
                    extra=vim.get('extra'),
                    deleted_at=datetime.min)
                context.session.add(vim_db)
                vim_auth_db = nfvo_db.VimAuth(
                    id=uuidutils.generate_uuid(),
                    vim_id=vim.get('id'),
                    password=vim_cred.pop('password'),
                    vim_project=vim.get('vim_project'),
                    auth_url=vim.get('auth_url'),
                    auth_cred=vim_cred)
                context.session.add(vim_auth_db)
        except DBDuplicateEntry as e:
            raise exceptions.DuplicateEntity(
                _type="vim",
                entry=e.columns)
        vim_dict = self._make_vim_dict(vim_db)
        self._cos_db_plg.create_event(
            context, res_id=vim_dict['id'],
            res_type=constants.RES_TYPE_VIM,
            res_state=vim_dict['status'],
            evt_type=constants.RES_EVT_CREATE,
            tstamp=vim_dict['created_at'])
        return vim_dict

    def delete_vim(self, context, vim_id, soft_delete=True):
        with context.session.begin(subtransactions=True):
            vim_db = self._get_resource(context, nfvo_db.Vim, vim_id)
            if soft_delete:
                vim_db.update({'deleted_at': timeutils.utcnow()})
                self._cos_db_plg.create_event(
                    context, res_id=vim_db['id'],
                    res_type=constants.RES_TYPE_VIM,
                    res_state=vim_db['status'],
                    evt_type=constants.RES_EVT_DELETE,
                    tstamp=vim_db[constants.RES_EVT_DELETED_FLD])
            else:
                context.session.query(nfvo_db.VimAuth).filter_by(
                    vim_id=vim_id).delete()
                context.session.delete(vim_db)

    def is_vim_still_in_use(self, context, vim_id):
        with context.session.begin(subtransactions=True):
            vnfs_db = self._model_query(context, vnfm_db.VNF).filter_by(
                vim_id=vim_id).first()
            if vnfs_db is not None:
                raise nfvo.VimInUseException(vim_id=vim_id)
        return vnfs_db

    def get_vim(self, context, vim_id, fields=None, mask_password=True):
        vim_db = self._get_resource(context, nfvo_db.Vim, vim_id)
        return self._make_vim_dict(vim_db, mask_password=mask_password)

    def get_vims(self, context, filters=None, fields=None):
        return self._get_collection(context, nfvo_db.Vim, self._make_vim_dict,
                                    filters=filters, fields=fields)

    def update_vim(self, context, vim_id, vim):
        self._validate_default_vim(context, vim, vim_id=vim_id)
        with context.session.begin(subtransactions=True):
            vim_cred = vim['auth_cred']
            vim_project = vim['vim_project']
            vim_db = self._get_resource(context, nfvo_db.Vim, vim_id)
            try:
                if 'name' in vim:
                    vim_db.update({'name': vim.get('name')})
                if 'description' in vim:
                    vim_db.update({'description': vim.get('description')})
                if 'is_default' in vim:
                    vim_db.update({'is_default': vim.get('is_default')})
                if 'placement_attr' in vim:
                    vim_db.update(
                        {'placement_attr': vim.get('placement_attr')})
                if 'extra' in vim:
                    vim_db.update({'extra': vim.get('extra')})
                vim_auth_db = (self._model_query(
                    context, nfvo_db.VimAuth).filter(
                    nfvo_db.VimAuth.vim_id == vim_id).with_for_update().one())
            except orm_exc.NoResultFound:
                raise nfvo.VimNotFoundException(vim_id=vim_id)
            vim_auth_db.update({'auth_cred': vim_cred, 'password':
                                vim_cred.pop('password'), 'vim_project':
                                vim_project})
            vim_db.update({'updated_at': timeutils.utcnow()})
            self._cos_db_plg.create_event(
                context, res_id=vim_db['id'],
                res_type=constants.RES_TYPE_VIM,
                res_state=vim_db['status'],
                evt_type=constants.RES_EVT_UPDATE,
                tstamp=vim_db[constants.RES_EVT_UPDATED_FLD])

        return self.get_vim(context, vim_id)

    def update_vim_status(self, context, vim_id, status):
        with context.session.begin(subtransactions=True):
            try:
                vim_db = (self._model_query(context, nfvo_db.Vim).filter(
                    nfvo_db.Vim.id == vim_id).with_for_update().one())
            except orm_exc.NoResultFound:
                raise nfvo.VimNotFoundException(vim_id=vim_id)
            vim_db.update({'status': status,
                           'updated_at': timeutils.utcnow()})
        return self._make_vim_dict(vim_db)

    def _validate_default_vim(self, context, vim, vim_id=None):
        if not vim.get('is_default'):
            return True
        try:
            vim_db = self._get_default_vim(context)
        except orm_exc.NoResultFound:
            return True
        if vim_id == vim_db.id:
            return True
        raise nfvo.VimDefaultDuplicateException(vim_id=vim_db.id)

    def _get_default_vim(self, context):
        query = self._model_query(context, nfvo_db.Vim)
        return query.filter(
            nfvo_db.Vim.tenant_id == context.tenant_id).filter(
            nfvo_db.Vim.is_default == sql.true()).one()

    def get_default_vim(self, context):
        vim_db = self._get_default_vim(context)
        return self._make_vim_dict(vim_db, mask_password=False)



# All Rights Reserved.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class ToscaKubeObject(object):

    """ToscaKubeObject holds the basic struct of a VDU.

    That is used for translating TOSCA to Kubernetes templates.
    """

    def __init__(self, name=None, namespace=None, mapping_ports=None,
                 containers=None, network_name=None,
                 mgmt_connection_point=False, scaling_object=None,
                 service_type=None, labels=None, annotations=None):
        self._name = name
        self._namespace = namespace
        self._mapping_ports = mapping_ports
        self._containers = containers
        self._network_name = network_name
        self._mgmt_connection_point = mgmt_connection_point
        self._scaling_object = scaling_object
        self._service_type = service_type
        self._labels = labels
        self._annotations = annotations

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    @property
    def namespace(self):
        return self._namespace

    @namespace.setter
    def namespace(self, namespace):
        self._namespace = namespace

    @property
    def mapping_ports(self):
        return self._mapping_ports

    @mapping_ports.setter
    def mapping_ports(self, mapping_ports):
        self._mapping_ports = mapping_ports

    @property
    def containers(self):
        return self._containers

    @containers.setter
    def containers(self, containers):
        self._containers = containers

    @property
    def network_name(self):
        return self._network_name

    @network_name.setter
    def network_name(self, network_name):
        self._network_name = network_name

    @property
    def mgmt_connection_point(self):
        return self._mgmt_connection_point

    @mgmt_connection_point.setter
    def mgmt_connection_point(self, mgmt_connection_point):
        self._mgmt_connection_point = mgmt_connection_point

    @property
    def scaling_object(self):
        return self._scaling_object

    @scaling_object.setter
    def scaling_object(self, scaling_object):
        self._scaling_object = scaling_object

    @property
    def service_type(self):
        return self._service_type

    @service_type.setter
    def service_type(self, service_type):
        self._service_type = service_type

    @property
    def labels(self):
        return self._labels

    @labels.setter
    def labels(self, labels):
        self._labels = labels

    @property
    def annotations(self):
        return self._annotations

    @annotations.setter
    def annotations(self, annotations):
        self._annotations = annotations


class Container(object):
    """Container holds the basic structs of a container"""
    def __init__(self, name=None, num_cpus=None, mem_size=None, image=None,
                 command=None, args=None, ports=None, config=None):
        self._name = name
        self._num_cpus = num_cpus
        self._mem_size = mem_size
        self._image = image
        self._command = command
        self._args = args
        self._ports = ports
        self._config = config

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    @property
    def num_cpus(self):
        return self._num_cpus

    @num_cpus.setter
    def num_cpus(self, num_cpus):
        self._num_cpus = num_cpus

    @property
    def mem_size(self):
        return self._mem_size

    @mem_size.setter
    def mem_size(self, mem_size):
        self._mem_size = mem_size

    @property
    def image(self):
        return self._image

    @image.setter
    def image(self, image):
        self._image = image

    @property
    def command(self):
        return self._command

    @command.setter
    def command(self, command):
        self._command = command

    @property
    def args(self):
        return self._args

    @args.setter
    def args(self, args):
        self._args = args

    @property
    def ports(self):
        return self._ports

    @ports.setter
    def ports(self, ports):
        self._ports = ports

    @property
    def config(self):
        return self._config

    @config.setter
    def config(self, config):
        self._config = config


class ScalingObject(object):
    """ScalingObject holds the basic struct of a horizontal pod auto-scaling"""
    def __init__(self, scaling_name=None, min_replicas=None, max_replicas=None,
                 scale_target_name=None,
                 target_cpu_utilization_percentage=None):
        self._scaling_name = scaling_name
        self._min_replicas = min_replicas
        self._max_replicas = max_replicas
        self._scale_target_name = scale_target_name
        self._target_cpu_utilization_percentage = \
            target_cpu_utilization_percentage

    @property
    def scaling_name(self):
        return self._scaling_name

    @scaling_name.setter
    def scaling_name(self, scaling_name):
        self._scaling_name = scaling_name

    @property
    def min_replicas(self):
        return self._min_replicas

    @min_replicas.setter
    def min_replicas(self, min_replicas):
        self._min_replicas = min_replicas

    @property
    def max_replicas(self):
        return self._max_replicas

    @max_replicas.setter
    def max_replicas(self, max_replicas):
        self._max_replicas = max_replicas

    @property
    def scale_target_name(self):
        return self._scale_target_name

    @scale_target_name.setter
    def scale_target_name(self, scale_target_name):
        self._scale_target_name = scale_target_name

    @property
    def target_cpu_utilization_percentage(self):
        return self._target_cpu_utilization_percentage

    @target_cpu_utilization_percentage.setter
    def target_cpu_utilization_percentage(self,
                                          target_cpu_utilization_percentage):
        self._target_cpu_utilization_percentage = \
            target_cpu_utilization_percentage



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from tacker.mistral import workflow_generator
from tacker.nfvo.workflows import vim_monitor

LOG = logging.getLogger(__name__)


class WorkflowGenerator(workflow_generator.WorkflowGeneratorBase):
    def __init__(self, vim_id, action):
        super(WorkflowGenerator, self).__init__(
            vim_monitor.RESOURCE_NAME, action)
        self.wf_identifier = 'vim_id_' + vim_id
        self._build_basic_workflow()

    def _add_ping_vim_tasks(self):
        task_dict = dict()
        task = self.wf_name + vim_monitor.PING_VIM_TASK_NAME
        task_dict[task] = {
            'action': 'tacker.vim_ping_action',
            'input': {'count': self.input_dict_data['count'],
                      'targetip': self.input_dict_data['targetip'],
                      'vim_id': self.input_dict_data['vim_id'],
                      'interval': self.input_dict_data['interval'],
                      'timeout': self.input_dict_data['timeout']},
        }
        return task_dict

    def get_input_dict(self):
        return self.input_dict

    def _build_input(self, vim_id, count, timeout,
                     interval, targetip):
        self.input_dict_data = {'vim_id': vim_id,
                                'count': count,
                                'timeout': timeout,
                                'interval': interval,
                                'targetip': targetip}
        self.input_dict[self.resource] = self.input_dict_data

    def monitor_ping_vim(self, vim_id=None, count=1, timeout=1,
                         interval=1, targetip="127.0.0.1"):
        self._build_input(vim_id, count, timeout,
                          interval, targetip)
        self.definition[self.wf_identifier]['tasks'] = dict()
        self.definition[self.wf_identifier]['tasks'].update(
            self._add_ping_vim_tasks())



# Copyright 2011 Citrix System.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from functools import cmp_to_key
from functools import wraps
import netaddr
from oslo_config import cfg
import oslo_i18n
from oslo_log import log as logging
from oslo_policy import policy as oslo_policy
from urllib import parse as urllib_parse
from webob import exc

from tacker._i18n import _
from tacker.common import constants
from tacker.common import exceptions
from tacker import wsgi

LOG = logging.getLogger(__name__)


def validate_supported_params(supported_params):
    """Decorator for Restful API methods which specifies supported parameters.

    If there are unsupported parameters in the request parameters, the http
    code 400 Bad Request will be returned and the user will be told which
    parameters in the request are not supported.
    """
    def decorator(f):
        @wraps(f)
        def wrapped(*args, **kwargs):
            request = kwargs.get('request')
            if not request:
                request = args[1]
            params = set(request.params.keys())
            unsupported_params = params - supported_params
            if unsupported_params:
                msg = _("Not supported parameters: %s") \
                    % ','.join(unsupported_params)
                raise exc.HTTPBadRequest(explanation=msg)
            return f(*args, **kwargs)
        return wrapped
    return decorator


def get_filters(request, attr_info, skips=None):
    """Extract the filters from the request string.

    Returns a dict of lists for the filters:
    check=a&check=b&name=Bob&
    becomes:
    {'check': ['a', 'b'], 'name': ['Bob']}
    """
    res = {}
    skips = skips or []
    for key, values in request.GET.dict_of_lists().items():
        if key in skips:
            continue
        values = [v for v in values if v]
        key_attr_info = attr_info.get(key, {})
        if 'convert_list_to' in key_attr_info:
            values = key_attr_info['convert_list_to'](values)
        elif 'convert_to' in key_attr_info:
            convert_to = key_attr_info['convert_to']
            values = [convert_to(v) for v in values]
        if values:
            res[key] = values
    return res


def get_previous_link(request, items, id_key):
    params = request.GET.copy()
    params.pop('marker', None)
    if items:
        marker = items[0][id_key]
        params['marker'] = marker
    params['page_reverse'] = True
    return "%s?%s" % (request.path_url, urllib_parse.urlencode(params))


def get_next_link(request, items, id_key):
    params = request.GET.copy()
    params.pop('marker', None)
    if items:
        marker = items[-1][id_key]
        params['marker'] = marker
    params.pop('page_reverse', None)
    return "%s?%s" % (request.path_url, urllib_parse.urlencode(params))


def get_limit_and_marker(request):
    """Return marker, limit tuple from request.

    :param request: `wsgi.Request` possibly containing 'marker' and 'limit'
                    GET variables. 'marker' is the id of the last element
                    the client has seen, and 'limit' is the maximum number
                    of items to return. If limit == 0, it means we needn't
                    pagination, then return None.
    """
    max_limit = _get_pagination_max_limit()
    limit = _get_limit_param(request)
    if max_limit > 0:
        limit = min(max_limit, limit) or max_limit
    if not limit:
        return None, None
    marker = request.GET.get('marker', None)
    return limit, marker


def _get_pagination_max_limit():
    max_limit = -1
    if (cfg.CONF.pagination_max_limit.lower() !=
            constants.PAGINATION_INFINITE):
        try:
            max_limit = int(cfg.CONF.pagination_max_limit)
            if max_limit == 0:
                raise ValueError()
        except ValueError:
            LOG.warning("Invalid value for pagination_max_limit: %s. It "
                        "should be an integer greater to 0",
                        cfg.CONF.pagination_max_limit)
    return max_limit


def _get_limit_param(request):
    """Extract integer limit from request or fail."""
    limit = -1
    try:
        limit = int(request.GET.get('limit', 0))
        if limit >= 0:
            return limit
    except ValueError:
        pass
    msg = _("Limit must be an integer 0 or greater and not '%d'") % limit
    raise exceptions.BadRequest(resource='limit', msg=msg)


def list_args(request, arg):
    """Extract the list of arg from request."""
    return [v for v in request.GET.getall(arg) if v]


def get_sorts(request, attr_info):
    """Extract sort_key and sort_dir from request.

    Return as: [(key1, value1), (key2, value2)]
    """
    sort_keys = list_args(request, "sort_key")
    sort_dirs = list_args(request, "sort_dir")
    if len(sort_keys) != len(sort_dirs):
        msg = _("The number of sort_keys and sort_dirs must be same")
        raise exc.HTTPBadRequest(explanation=msg)
    valid_dirs = [constants.SORT_DIRECTION_ASC, constants.SORT_DIRECTION_DESC]
    absent_keys = [x for x in sort_keys if x not in attr_info]
    if absent_keys:
        msg = _("%s is invalid attribute for sort_keys") % absent_keys
        raise exc.HTTPBadRequest(explanation=msg)
    invalid_dirs = [x for x in sort_dirs if x not in valid_dirs]
    if invalid_dirs:
        msg = (_("%(invalid_dirs)s is invalid value for sort_dirs, "
                 "valid value is '%(asc)s' and '%(desc)s'") %
               {'invalid_dirs': invalid_dirs,
                'asc': constants.SORT_DIRECTION_ASC,
                'desc': constants.SORT_DIRECTION_DESC})
        raise exc.HTTPBadRequest(explanation=msg)
    return list(zip(sort_keys,
                    [x == constants.SORT_DIRECTION_ASC for x in sort_dirs]))


def get_page_reverse(request):
    data = request.GET.get('page_reverse', 'False')
    return data.lower() == "true"


def get_pagination_links(request, items, limit,
                         marker, page_reverse, key="id"):
    key = key if key else 'id'
    links = []
    if not limit:
        return links
    if not (len(items) < limit and not page_reverse):
        links.append({"rel": "next",
                      "href": get_next_link(request, items,
                                            key)})
    if not (len(items) < limit and page_reverse):
        links.append({"rel": "previous",
                      "href": get_previous_link(request, items,
                                                key)})
    return links


class PaginationHelper(object):

    def __init__(self, request, primary_key='id'):
        self.request = request
        self.primary_key = primary_key

    def update_fields(self, original_fields, fields_to_add):
        pass

    def update_args(self, args):
        pass

    def paginate(self, items):
        return items

    def get_links(self, items):
        return {}


class PaginationEmulatedHelper(PaginationHelper):

    def __init__(self, request, primary_key='id'):
        super(PaginationEmulatedHelper, self).__init__(request, primary_key)
        self.limit, self.marker = get_limit_and_marker(request)
        self.page_reverse = get_page_reverse(request)

    def update_fields(self, original_fields, fields_to_add):
        if not original_fields:
            return
        if self.primary_key not in original_fields:
            original_fields.append(self.primary_key)
            fields_to_add.append(self.primary_key)

    def paginate(self, items):
        if not self.limit:
            return items
        i = -1
        if self.marker:
            for item in items:
                i = i + 1
                if item[self.primary_key] == self.marker:
                    break
        if self.page_reverse:
            return items[i - self.limit:i]
        return items[i + 1:i + self.limit + 1]

    def get_links(self, items):
        return get_pagination_links(
            self.request, items, self.limit, self.marker,
            self.page_reverse, self.primary_key)


class PaginationNativeHelper(PaginationEmulatedHelper):

    def update_args(self, args):
        if self.primary_key not in dict(args.get('sorts', [])):
            args.setdefault('sorts', []).append((self.primary_key, True))
        args.update({'limit': self.limit, 'marker': self.marker,
                     'page_reverse': self.page_reverse})

    def paginate(self, items):
        return items


class NoPaginationHelper(PaginationHelper):
    pass


class SortingHelper(object):

    def __init__(self, request, attr_info):
        pass

    def update_args(self, args):
        pass

    def update_fields(self, original_fields, fields_to_add):
        pass

    def sort(self, items):
        return items


class SortingEmulatedHelper(SortingHelper):

    def __init__(self, request, attr_info):
        super(SortingEmulatedHelper, self).__init__(request, attr_info)
        self.sort_dict = get_sorts(request, attr_info)

    def update_fields(self, original_fields, fields_to_add):
        if not original_fields:
            return
        for key in dict(self.sort_dict):
            if key not in original_fields:
                original_fields.append(key)
                fields_to_add.append(key)

    def sort(self, items):
        def cmp_func(obj1, obj2):
            for key, direction in self.sort_dict:
                o1 = obj1[key]
                o2 = obj2[key]

                if o1 is None and o2 is None:
                    ret = 0
                elif o1 is None and o2 is not None:
                    ret = -1
                elif o1 is not None and o2 is None:
                    ret = 1
                else:
                    ret = (o1 > o2) - (o1 < o2)
                if ret:
                    return ret * (1 if direction else -1)
            return 0
        return sorted(items, key=cmp_to_key(cmp_func))


class SortingNativeHelper(SortingHelper):

    def __init__(self, request, attr_info):
        super(SortingNativeHelper, self).__init__(request, attr_info)
        self.sort_dict = get_sorts(request, attr_info)

    def update_args(self, args):
        args['sorts'] = self.sort_dict


class NoSortingHelper(SortingHelper):
    pass


class TackerController(object):
    """Base controller class for Tacker API."""

    # _resource_name will be redefined in sub concrete controller
    _resource_name = None

    def __init__(self, plugin):
        self._plugin = plugin
        super(TackerController, self).__init__()

    def _prepare_request_body(self, body, params):
        """Verify required parameters are in request body.

        Sets default value for missing optional parameters.
        Body argument must be the deserialized body.
        """
        try:
            if body is None:
                # Initialize empty resource for setting default value
                body = {self._resource_name: {}}
            data = body[self._resource_name]
        except KeyError:
            # raise if _resource_name is not in req body.
            raise exc.HTTPBadRequest(_("Unable to find '%s' in request body") %
                                     self._resource_name)
        for param in params:
            param_name = param['param-name']
            param_value = data.get(param_name)
            # If the parameter wasn't found and it was required, return 400
            if param_value is None and param['required']:
                msg = (_("Failed to parse request. "
                         "Parameter '%s' not specified") % param_name)
                LOG.error(msg)
                raise exc.HTTPBadRequest(msg)
            data[param_name] = param_value or param.get('default-value')
        return body


def convert_exception_to_http_exc(e, faults, language):
    serializer = wsgi.JSONDictSerializer()
    e = translate(e, language)
    body = serializer.serialize(
        {'TackerError': get_exception_data(e)})
    kwargs = {'body': body, 'content_type': 'application/json'}
    if isinstance(e, exc.HTTPException):
        # already an HTTP error, just update with content type and body
        e.body = body
        e.content_type = kwargs['content_type']
        return e
    if isinstance(e, (exceptions.TackerException, netaddr.AddrFormatError,
                      oslo_policy.PolicyNotAuthorized)):
        for fault in faults:
            if isinstance(e, fault):
                mapped_exc = faults[fault]
                break
        else:
            mapped_exc = exc.HTTPInternalServerError
        return mapped_exc(**kwargs)
    if isinstance(e, NotImplementedError):
        # NOTE(armando-migliaccio): from a client standpoint
        # it makes sense to receive these errors, because
        # extensions may or may not be implemented by
        # the underlying plugin. So if something goes south,
        # because a plugin does not implement a feature,
        # returning 500 is definitely confusing.
        kwargs['body'] = serializer.serialize(
            {'NotImplementedError': get_exception_data(e)})
        return exc.HTTPNotImplemented(**kwargs)
    # NOTE(jkoelker) Everything else is 500
    # Do not expose details of 500 error to clients.
    msg = _('Request Failed: internal server error while '
            'processing your request.')
    msg = translate(msg, language)
    kwargs['body'] = serializer.serialize(
        {'TackerError': get_exception_data(exc.HTTPInternalServerError(msg))})
    return exc.HTTPInternalServerError(**kwargs)


def get_exception_data(e):
    """Extract the information about an exception.

    Tacker client for the v1 API expects exceptions to have 'type', 'message'
    and 'detail' attributes.This information is extracted and converted into a
    dictionary.

    :param e: the exception to be reraised
    :returns: a structured dict with the exception data
    """
    err_data = {'type': e.__class__.__name__,
                'message': e, 'detail': ''}
    return err_data


def translate(translatable, locale):
    """Translate the object to the given locale.

    If the object is an exception its translatable elements are translated
    in place, if the object is a translatable string it is translated and
    returned. Otherwise, the object is returned as-is.

    :param translatable: the object to be translated
    :param locale: the locale to translate to
    :returns: the translated object, or the object as-is if it
              was not translated
    """
    localize = oslo_i18n.translate
    if isinstance(translatable, exceptions.TackerException):
        translatable.msg = localize(translatable.msg, locale)
    elif isinstance(translatable, exc.HTTPError):
        translatable.detail = localize(translatable.detail, locale)
    elif isinstance(translatable, Exception):
        translatable.message = localize(translatable, locale)
    else:
        return localize(translatable, locale)
    return translatable



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import log as logging

from tacker.common import csar_utils
from tacker.common import exceptions
from tacker.common import utils
from tacker.conductor.conductorrpc import vnf_pkgm_rpc
from tacker.glance_store import store as glance_store
from tacker import objects
from tacker.objects import fields
import tacker.vnfm.nfvo_client as nfvo_client
import time
import webob


CONF = cfg.CONF

LOG = logging.getLogger(__name__)


class SyncVnfPackage:

    vnf_package_rpc_api = vnf_pkgm_rpc.VNFPackageRPCAPI()

    @classmethod
    def create_package(cls, context, vnf_package_info):
        """vnf_package, create a vnf_package_vnfd table."""

        vnf_package_info = utils.convert_camelcase_to_snakecase(
            vnf_package_info)

        try:
            vnf_package = cls.__create_vnf_package(context, vnf_package_info)
        except Exception as exc:
            raise webob.exc.HTTPInternalServerError(
                explanation=exc)

        try:
            artifact_paths = cls._get_artifact_paths(vnf_package_info)
            vnf_package_binary = \
                nfvo_client.VnfPackageRequest.download_vnf_packages(
                    vnf_package.id, artifact_paths)
        except nfvo_client.UndefinedExternalSettingException as exc:
            raise webob.exc.HTTPNotFound(explanation=exc)
        except (nfvo_client.FaliedDownloadContentException, Exception) as exc:
            raise webob.exc.HTTPInternalServerError(
                explanation=exc)

        try:
            (location, size, _, multihash, _) = glance_store.store_csar(
                context, vnf_package.id, vnf_package_binary)

            cls.__update_vnf_package(vnf_package, location, size, multihash)
            cls.vnf_package_rpc_api.upload_vnf_package_content(
                context, vnf_package)
            vnf_package_vnfd = cls._get_vnf_package_vnfd(
                context, vnf_package_info.get('vnfd_id'))
        except Exception as exc:
            raise webob.exc.HTTPInternalServerError(
                explanation=exc)

        return vnf_package_vnfd

    @classmethod
    def _get_artifact_paths(cls, vnf_package_info):
        additional_artifacts = vnf_package_info.get('additional_artifacts')
        if additional_artifacts is None:
            return None

        return [artifact.get('artifact_path')
                for artifact in additional_artifacts
                if 'artifact_path' in artifact]

    @classmethod
    def __store_csar(cls, context, id, body):
        (location, size, checksum, multihash,
         loc_meta) = glance_store.store_csar(context, id, body)
        return location, size, checksum, multihash, loc_meta

    @classmethod
    def __load_csar(cls, context, vnf_package):
        location = vnf_package.location_glance_store
        zip_path = glance_store.load_csar(vnf_package.id, location)
        vnf_data, flavours = csar_utils.load_csar_data(
            context.elevated(), vnf_package.id, zip_path)
        return vnf_data, flavours

    @classmethod
    def __create_vnf_package(cls, context, vnf_package_info):
        """VNF Package Table Registration."""
        vnf_package = objects.VnfPackage(
            context=context,
            id=vnf_package_info.get('id'),
            onboarding_state=fields.PackageOnboardingStateType.CREATED,
            operational_state=fields.PackageOperationalStateType.DISABLED,
            usage_state=fields.PackageUsageStateType.NOT_IN_USE,
            tenant_id=context.project_id
        )
        vnf_package.create()
        return vnf_package

    @classmethod
    def __update_vnf_package(cls, vnf_package, location, size, multihash):
        """VNF Package Table Update."""
        vnf_package.algorithm = CONF.vnf_package.hashing_algorithm
        vnf_package.location_glance_store = location
        vnf_package.hash = multihash
        vnf_package.size = size
        vnf_package.save()

    @classmethod
    def _get_vnf_package_vnfd(cls, context, vnfd_id):
        """Get VNF Package VNFD."""
        for num in range(CONF.vnf_lcm.retry_num):
            try:
                vnfd = objects.VnfPackageVnfd.get_by_id(
                    context,
                    vnfd_id)
                return vnfd
            except exceptions.VnfPackageVnfdNotFound:
                LOG.debug("retry_wait %s" %
                    CONF.vnf_lcm.retry_wait)
                time.sleep(CONF.vnf_lcm.retry_wait)

        return None



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from tacker import context
from tacker.sol_refactored.api import api_version
from tacker.sol_refactored.api import validator
from tacker.sol_refactored.common import exceptions as sol_ex
from tacker.tests import base


test_schema_v200 = {
    'type': 'object',
    'properties': {
        'vnfdId': {'type': 'string'},
        'ProductId': {'type': 'string'}
    },
    'required': ['vnfdId', 'ProductId'],
    'additionalProperties': True
}

test_schema_v210 = {
    'type': 'object',
    'properties': {
        'vnfdId': {'type': 'string'},
        'flavourId': {'type': 'string'}
    },
    'required': ['vnfdId', 'flavourId'],
    'additionalProperties': True
}


class TestValidator(base.BaseTestCase):

    def setUp(self):
        super(TestValidator, self).setUp()
        self.context = context.get_admin_context_without_session()
        self.request = mock.Mock()
        self.request.context = self.context

    @validator.schema(test_schema_v200, '2.0.0', '2.0.2')
    @validator.schema(test_schema_v210, '2.1.0', '2.2.0')
    def _test_method(self, request, body):
        return True

    @mock.patch.object(api_version, 'supported_versions',
        new=['2.0.0', '2.0.1', '2.0.2', '2.1.0', '2.2.0'])
    def test_validator(self):
        body = {"vnfdId": "vnfd_id", "ProductId": "product_id"}
        for ok_ver in ['2.0.0', '2.0.1', '2.0.2']:
            self.context.api_version = api_version.APIVersion(ok_ver)
            result = self._test_method(request=self.request, body=body)
            self.assertTrue(result)
        for ng_ver in ['2.1.0', '2.2.0']:
            self.context.api_version = api_version.APIVersion(ng_ver)
            self.assertRaises(sol_ex.SolValidationError,
                self._test_method, request=self.request, body=body)

        body = {"vnfdId": "vnfd_id", "flavourId": "flavour_id"}
        for ok_ver in ['2.1.0', '2.2.0']:
            self.context.api_version = api_version.APIVersion(ok_ver)
            result = self._test_method(request=self.request, body=body)
            self.assertTrue(result)
        for ng_ver in ['2.0.0', '2.0.1', '2.0.2']:
            self.context.api_version = api_version.APIVersion(ng_ver)
            self.assertRaises(sol_ex.SolValidationError,
                self._test_method, request=self.request, body=body)



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from tacker.sol_refactored.api import api_version
from tacker.sol_refactored.api import wsgi as sol_wsgi


class VnfLcmVersionsController(sol_wsgi.SolAPIController):
    def index(self, request):
        api_versions = (api_version.supported_versions_v1['apiVersions'] +
                        api_version.supported_versions_v2['apiVersions'])
        body = {"uriPrefix": "/vnflcm",
                "apiVersions": api_versions}
        return sol_wsgi.SolResponse(200, body)



# Copyright 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import testtools

from tacker.tests import base
from tacker import wsgi


class ExpectedException(testtools.ExpectedException):
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if super(ExpectedException, self).__exit__(exc_type,
                                                   exc_value,
                                                   traceback):
            self.exception = exc_value
            return True
        return False


def create_request(path, body, content_type, method='GET',
                   query_string=None, context=None):
    if query_string:
        url = "%s?%s" % (path, query_string)
    else:
        url = path
    req = wsgi.Request.blank(url)
    req.method = method
    req.headers = {}
    req.headers['Accept'] = content_type
    req.body = body
    if context:
        req.environ['tacker.context'] = context
    return req


class WebTestCase(base.BaseTestCase):
    fmt = 'json'

    def setUp(self):
        super(WebTestCase, self).setUp()
        json_deserializer = wsgi.JSONDeserializer()
        self._deserializers = {
            'application/json': json_deserializer,
        }

    def deserialize(self, response):
        ctype = 'application/%s' % self.fmt
        data = self._deserializers[ctype].deserialize(response.body)['body']
        return data

    def serialize(self, data):
        ctype = 'application/%s' % self.fmt
        result = wsgi.Serializer().serialize(data, ctype)
        return result


class SubDictMatch(object):

    def __init__(self, sub_dict):
        self.sub_dict = sub_dict

    def __eq__(self, super_dict):
        return all(item in super_dict.items()
                   for item in self.sub_dict.items())

    def __ne__(self, super_dict):
        return not self.__eq__(super_dict)



# Copyright (C) 2019 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""add db tables for vnf packages

Revision ID: 9d425296f2c3
Revises: cd04a8335c18
Create Date: 2019-06-03 08:37:05.095587

"""

# flake8: noqa: E402

# revision identifiers, used by Alembic.
revision = '9d425296f2c3'
down_revision = 'cd04a8335c18'

from alembic import op
import sqlalchemy as sa
from sqlalchemy import Boolean

from tacker.db import types


def upgrade(active_plugins=None, options=None):
    op.create_table(
        'vnf_packages',
        sa.Column('id', types.Uuid(length=36), nullable=False),
        sa.Column('onboarding_state', sa.String(length=255), nullable=False),
        sa.Column('operational_state', sa.String(length=255), nullable=False),
        sa.Column('usage_state', sa.String(length=255), nullable=False),
        sa.Column('tenant_id', sa.String(length=64), nullable=False),
        sa.Column('algorithm', sa.String(length=64), nullable=True),
        sa.Column('hash', sa.String(length=128), nullable=True),
        sa.Column('location_glance_store', sa.Text(), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', Boolean, default=False),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB'
    )
    op.create_table(
        'vnf_packages_user_data',
        sa.Column('id', sa.Integer, nullable=False, autoincrement=True),
        sa.Column('package_uuid', types.Uuid(length=36), nullable=False),
        sa.Column('key', sa.String(length=255), nullable=False),
        sa.Column('value', sa.String(length=255), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', Boolean, default=False),
        sa.PrimaryKeyConstraint('id'),
        sa.ForeignKeyConstraint(['package_uuid'],
                                ['vnf_packages.id'], ),
        sa.Index('vnf_packages_user_data_key_idx', 'key'),
        sa.Index('vnf_packages_user_data_value_idx', 'value'),
        sa.UniqueConstraint('id', 'key', 'deleted',
                name='uniq_vnf_packages_user_data0idid0key0deleted'),
        mysql_engine='InnoDB'
    )
    op.create_table(
        'vnf_package_vnfd',
        sa.Column('id', types.Uuid(length=36), nullable=False),
        sa.Column('package_uuid', types.Uuid(length=36), nullable=False),
        sa.Column('vnfd_id', types.Uuid(length=36), nullable=False),
        sa.Column('vnf_provider', sa.String(length=255), nullable=False),
        sa.Column('vnf_product_name', sa.String(length=255), nullable=False),
        sa.Column('vnf_software_version', sa.String(length=255),
                  nullable=False),
        sa.Column('vnfd_version', sa.String(length=255), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', Boolean, default=False),
        sa.PrimaryKeyConstraint('id'),
        sa.ForeignKeyConstraint(['package_uuid'],
                                ['vnf_packages.id'], ),
        mysql_engine='InnoDB'
    )
    op.create_table(
        'vnf_deployment_flavours',
        sa.Column('id', types.Uuid(length=36), nullable=False),
        sa.Column('package_uuid', types.Uuid(length=36), nullable=False),
        sa.Column('flavour_id', sa.String(length=255), nullable=False),
        sa.Column('flavour_description', sa.Text(), nullable=False),
        sa.Column('instantiation_levels', sa.Text(), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', Boolean, default=False),
        sa.PrimaryKeyConstraint('id'),
        sa.ForeignKeyConstraint(['package_uuid'],
                                ['vnf_packages.id'], ),
        mysql_engine='InnoDB'
    )
    op.create_table(
        'vnf_software_images',
        sa.Column('id', types.Uuid(length=36), nullable=False),
        sa.Column('software_image_id', sa.String(length=255), nullable=False),
        sa.Column('flavour_uuid', types.Uuid(length=36), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=False),
        sa.Column('provider', sa.String(length=255), nullable=False),
        sa.Column('version', sa.String(length=255), nullable=False),
        sa.Column('algorithm', sa.String(length=64), nullable=False),
        sa.Column('hash', sa.String(length=128), nullable=False),
        sa.Column('container_format', sa.String(length=20), nullable=False),
        sa.Column('disk_format', sa.String(length=20), nullable=False),
        sa.Column('min_disk', sa.Integer, nullable=False),
        sa.Column('min_ram', sa.Integer, nullable=False),
        sa.Column('size', sa.BigInteger, nullable=False),
        sa.Column('image_path', sa.Text(), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', Boolean, default=False),
        sa.PrimaryKeyConstraint('id'),
        sa.ForeignKeyConstraint(['flavour_uuid'],
                                ['vnf_deployment_flavours.id'], ),
        mysql_engine='InnoDB'
    )
    op.create_table(
        'vnf_software_image_metadata',
        sa.Column('id', sa.Integer, nullable=False, autoincrement=True),
        sa.Column('image_uuid', types.Uuid(length=36), nullable=False),
        sa.Column('key', sa.String(length=255), nullable=False),
        sa.Column('value', sa.String(length=255), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', Boolean, default=False),
        sa.PrimaryKeyConstraint('id'),
        sa.ForeignKeyConstraint(['image_uuid'],
                                ['vnf_software_images.id'], ),
        sa.Index('vnf_software_image_metadata_key_idx', 'key'),
        sa.Index('vnf_software_image_metadata_value_idx', 'value'),
        sa.UniqueConstraint('id', 'key', 'deleted',
                name='uniq_vnf_software_image_metadata0idid0key0deleted'),
        mysql_engine='InnoDB'
    )



# All Rights Reserved.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from tacker.api import extensions


class VimAbstractDriver(extensions.PluginInterface, metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def get_type(self):
        """Get VIM Driver type

        Return one of predefined types of VIMs.
        """
        pass

    @abc.abstractmethod
    def get_name(self):
        """Get VIM name

        Return a symbolic name for the VIM driver.
        """
        pass

    @abc.abstractmethod
    def get_description(self):
        pass

    @abc.abstractmethod
    def register_vim(self, vim_obj):
        """Register VIM object in to NFVO plugin

        Validate, encode and store VIM information for deploying VNFs.
        """
        pass

    @abc.abstractmethod
    def deregister_vim(self, vim_obj):
        """Deregister VIM object from NFVO plugin

        Cleanup VIM data and delete VIM information
        """
        pass

    @abc.abstractmethod
    def authenticate_vim(self, vim_obj):
        """Authenticate VIM connection parameters

        Validate authentication credentials and connectivity of VIM
        """
        pass

    @abc.abstractmethod
    def encode_vim_auth(self, vim_id, auth):
        """Encrypt VIM credentials

        Encrypt and store VIM sensitive information such as password
        """
        pass

    @abc.abstractmethod
    def delete_vim_auth(self, vim_id, auth):
        """Delete VIM auth keys

        Delete VIM sensitive information such as keys from file system or DB
        """
        pass

    @abc.abstractmethod
    def get_vim_resource_id(self, vim_obj, resource_type, resource_name):
        """Parses a VIM resource ID from a given type and name

        :param vim_obj: VIM information
        :param resource_type: type of resource, such as network, compute
        :param resource_name: name of resource, such at "test-network"
        :return: ID of resource
        """
        pass



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import hashlib
import os

from tacker.tests.functional.sol.vnflcm import base as vnflcm_base
from tacker.tests.functional.sol.vnflcm import fake_vnflcm
from tacker.tests.functional.sol_separated_nfvo.vnflcm import fake_grant
from tacker.tests.functional.sol_separated_nfvo.vnflcm import fake_vnfpkgm


class VnfLcmWithNfvoSeparator(vnflcm_base.BaseVnfLcmTest):

    def _register_vnf_package_mock_response(self, package_dir="functional6"):
        """Prepare VNF package for test.

        Register VNF package response to fake NFVO server and Cleanups.

        Returns:
            Response: VNF Package information
        """
        # Pre Setting: Create vnf package.
        sample_name = package_dir
        csar_package_path = os.path.abspath(
            os.path.join(
                os.path.dirname(__file__),
                "../../../etc/samples/etsi/nfv",
                sample_name))

        # Get VNFD id.
        tempname, vnfd_id = vnflcm_base._create_csar_with_unique_vnfd_id(
            csar_package_path)
        with open(tempname, "rb") as f:
            vnf_package_hash = hashlib.sha256(f.read()).hexdigest()

        vnf_package_info = \
            fake_vnfpkgm.VnfPackage.make_individual_response_body(
                vnfd_id, vnf_package_hash)
        vnf_package_id = vnf_package_info['id']

        # Post Setting: Reserve deleting vnf package.
        self.addCleanup(vnflcm_base._delete_vnf_package, self.tacker_client,
            vnf_package_id)

        # Set "VNF Packages" response
        vnflcm_base.FAKE_SERVER_MANAGER.set_callback('GET',
            fake_vnfpkgm.VnfPackage.VNF_PACKAGE_REQ_PATH, status_code=200,
            response_body=[vnf_package_info])

        # Set "VNF Package content" response
        vnflcm_base.FAKE_SERVER_MANAGER.set_callback('GET',
            os.path.join(
                fake_vnfpkgm.VnfPackage.VNF_PACKAGE_REQ_PATH,
                vnf_package_id,
                'package_content'),
            status_code=200,
            response_headers={"Content-Type": "application/zip"},
            content=tempname)

        # Set "Individual VNF package artifact" response
        vnflcm_base.FAKE_SERVER_MANAGER.set_callback('GET',
            os.path.join(
                fake_vnfpkgm.VnfPackage.VNF_PACKAGE_REQ_PATH,
                vnf_package_id,
                'artifacts',
                vnf_package_info['additionalArtifacts'][0]['artifactPath']),
            status_code=200,
            response_headers={"Content-Type": "application/zip"},
            content=tempname)

        # Set "VNFD of individual VNF package" response
        vnflcm_base.FAKE_SERVER_MANAGER.set_callback('GET',
            os.path.join(
                fake_vnfpkgm.VnfPackage.VNF_PACKAGE_REQ_PATH,
                vnf_package_id,
                'vnfd'),
            status_code=200,
            response_headers={"Content-Type": "application/zip"},
            content=tempname)

        return vnf_package_info

    def test_inst_chgextconn_term(self):
        """Test basic life cycle operations with sample VNFD with UserData.

        In this test case, we do following steps.
            - Create subscription.
            - Test notification.
            - Create VNF instance.
            - Instantiate VNF.
            - List VNF instances.
            - Show VNF instance.
            - Change External VNF Connectivity.
            - Get opOccs information.
            - Terminate VNF.
            - Delete VNF.
            - Delete subscription.
            - Show subscription.
        """
        vnf_package_info = self._register_vnf_package_mock_response()
        glance_image = fake_grant.Grant.get_sw_image()
        flavour_vdu_dict = fake_grant.Grant.get_compute_flavor()
        zone_name_list = []
        availability_zone_info = self._list_zone()
        zone_name_list = list(set(
            [zone.zone for zone in availability_zone_info
             if zone.binary == 'nova-compute']))
        # Create subscription and register it.
        callback_url = os.path.join(vnflcm_base.MOCK_NOTIFY_CALLBACK_URL,
            self._testMethodName)
        request_body = fake_vnflcm.Subscription.make_create_request_body(
            'http://localhost:{}{}'.format(
                vnflcm_base.FAKE_SERVER_MANAGER.SERVER_PORT,
                callback_url))
        resp, response_body = self._register_subscription(request_body)
        self.assertEqual(201, resp.status_code)
        self.assert_http_header_location_for_subscription(resp.headers)
        subscription_id = response_body.get('id')
        self.addCleanup(self._delete_subscription, subscription_id)

        # Test notification
        self.assert_notification_get(callback_url)

        # Create vnf instance
        resp, vnf_instance = self._create_vnf_instance_from_body(
            fake_vnflcm.VnfInstances.make_create_request_body(
                vnf_package_info['vnfdId']))
        vnf_instance_id = vnf_instance.get('id')
        self._wait_lcm_done(vnf_instance_id=vnf_instance_id)
        self._assert_create_vnf(resp, vnf_instance)
        vnf_instance_name = vnf_instance['vnfInstanceName']
        self.addCleanup(self._delete_vnf_instance, vnf_instance_id)

        # Set Fake server response for Grant-Req(Instantiate)
        vnflcm_base.FAKE_SERVER_MANAGER.set_callback('POST',
            fake_grant.Grant.GRANT_REQ_PATH, status_code=201,
            callback=lambda req_headers,
            req_body: fake_grant.Grant.make_inst_response_body(
                req_body,
                self.vim['tenant_id'], glance_image,
                flavour_vdu_dict,
                zone_name_list))

        # Instantiate vnf instance
        request_body = fake_vnflcm.VnfInstances.\
            make_inst_request_body_include_num_dynamic(
                self.vim['tenant_id'], self.ext_networks,
                self.ext_mngd_networks, self.ext_link_ports, self.ext_subnets)
        resp, _ = self._instantiate_vnf_instance(vnf_instance_id, request_body)
        self._wait_lcm_done('COMPLETED', vnf_instance_id=vnf_instance_id)
        self._assert_instantiate_vnf(resp, vnf_instance_id)

        # List vnf instances
        filter_expr = {
            'filter': "(eq,id,{});(eq,vnfInstanceName,{})".format(
                vnf_instance_id, vnf_instance_name)}
        resp, vnf_instances = self._list_vnf_instance(params=filter_expr)
        self.assertEqual(200, resp.status_code)
        self.assertEqual(1, len(vnf_instances))

        # Show vnf instance
        resp, vnf_instance = self._show_vnf_instance(vnf_instance_id)
        self.assertEqual(200, resp.status_code)

        # Set Fake server response for Grant-Req(Chnage-ext-conn)
        vnflcm_base.FAKE_SERVER_MANAGER.set_callback(
            'POST',
            fake_grant.Grant.GRANT_REQ_PATH,
            status_code=201,
            callback=lambda req_headers,
            req_body: fake_grant.Grant.make_change_ext_conn_response_body(
                req_body,
                self.vim['tenant_id'],
                zone_name_list))

        # Change external connectivity
        request_body = \
            fake_vnflcm.VnfInstances.make_change_ext_conn_request_body(
                self.vim['tenant_id'], self.changed_ext_networks,
                self.changed_ext_subnets)
        before_fixed_ips = self._get_fixed_ips(vnf_instance_id, request_body)
        resp, _ = \
            self._change_ext_conn_vnf_instance(vnf_instance_id, request_body)
        self._wait_lcm_done('COMPLETED', vnf_instance_id=vnf_instance_id)
        after_fixed_ips = self._get_fixed_ips(vnf_instance_id, request_body)
        self.assertNotEqual(before_fixed_ips, after_fixed_ips)
        vnflcm_base.FAKE_SERVER_MANAGER.clear_history(
            fake_grant.Grant.GRANT_REQ_PATH)

        # get vnflcm_op_occ_id
        callback_url = os.path.join(
            vnflcm_base.MOCK_NOTIFY_CALLBACK_URL,
            self._testMethodName)
        notify_mock_responses = vnflcm_base.FAKE_SERVER_MANAGER.get_history(
            callback_url)
        vnflcm_base.FAKE_SERVER_MANAGER.clear_history(
            callback_url)

        vnflcm_op_occ_id = notify_mock_responses[0].request_body.get(
            'vnfLcmOpOccId')
        self.assertIsNotNone(vnflcm_op_occ_id)

        # occ-show(chgextconn)
        resp, op_occs_info = self._show_op_occs(vnflcm_op_occ_id)
        self._assert_occ_show(resp, op_occs_info)

        # Set Fake server response for Grant-Req(Terminate)
        vnflcm_base.FAKE_SERVER_MANAGER.set_callback('POST',
            fake_grant.Grant.GRANT_REQ_PATH, status_code=201,
            callback=lambda req_headers,
            req_body: fake_grant.Grant.make_term_response_body(req_body))

        # Get stack informations to terminate.
        stack = self._get_heat_stack(vnf_instance_id)
        resources_list = self._get_heat_resource_list(stack.id)
        resource_name_list = [r.resource_name for r in resources_list]
        glance_image_id_list = self._get_glance_image_list_from_stack_resource(
            stack.id, resource_name_list)

        # Terminate VNF
        terminate_req_body = fake_vnflcm.VnfInstances.make_term_request_body()
        resp, _ = self._terminate_vnf_instance(vnf_instance_id,
                terminate_req_body)
        self._wait_lcm_done('COMPLETED', vnf_instance_id=vnf_instance_id)
        self._assert_terminate_vnf(resp, vnf_instance_id, stack.id,
             resource_name_list, glance_image_id_list)

        # Delete VNF
        resp, _ = self._delete_vnf_instance(vnf_instance_id)
        self._wait_lcm_done(vnf_instance_id=vnf_instance_id)
        self.assert_delete_vnf(resp, vnf_instance_id)

        # Delete Subscription
        resp, response_body = self._delete_subscription(subscription_id)
        self.assertEqual(204, resp.status_code)

        # Check subscription was deleted
        resp, show_body = self._show_subscription(subscription_id)
        self.assertEqual(404, resp.status_code)

    def test_inst_heal_term(self):
        """Test basic life cycle operations with sample VNFD with UserData.

        In this test case, we do following steps.
            - Create subscription.
            - Create VNF instance.
            - Instantiate VNF.
            - Heal VNF with all VNFc.
            - Terminate VNF
            - Delete VNF
            - Delete subscription
        """
        vnf_package_info = self._register_vnf_package_mock_response()
        glance_image = fake_grant.Grant.get_sw_image()
        flavour_vdu_dict = fake_grant.Grant.get_compute_flavor()
        zone_name_list = []
        availability_zone_info = self._list_zone()
        zone_name_list = list(set(
            [zone.zone for zone in availability_zone_info
             if zone.binary == 'nova-compute']))

        # Create subscription and register it.
        callback_url = os.path.join(vnflcm_base.MOCK_NOTIFY_CALLBACK_URL,
            self._testMethodName)
        request_body = fake_vnflcm.Subscription.make_create_request_body(
            'http://localhost:{}{}'.format(
                vnflcm_base.FAKE_SERVER_MANAGER.SERVER_PORT,
                callback_url))
        resp, response_body = self._register_subscription(request_body)
        self.assertEqual(201, resp.status_code)
        self.assert_http_header_location_for_subscription(resp.headers)
        self.assert_notification_get(callback_url)
        subscription_id = response_body.get('id')
        self.addCleanup(self._delete_subscription, subscription_id)

        # Create vnf instance
        resp, vnf_instance = self._create_vnf_instance_from_body(
            fake_vnflcm.VnfInstances.make_create_request_body(
                vnf_package_info['vnfdId']))
        vnf_instance_id = vnf_instance.get('id')
        self._wait_lcm_done(vnf_instance_id=vnf_instance_id)
        self._assert_create_vnf(resp, vnf_instance)
        self.addCleanup(self._delete_vnf_instance, vnf_instance_id)

        # Set Fake server response for Grant-Req(Instantiate)
        vnflcm_base.FAKE_SERVER_MANAGER.set_callback('POST',
            fake_grant.Grant.GRANT_REQ_PATH, status_code=201,
            callback=lambda req_headers,
            req_body: fake_grant.Grant.make_inst_response_body(
                req_body,
                self.vim['tenant_id'],
                glance_image,
                flavour_vdu_dict,
                zone_name_list))

        # Instantiate vnf instance
        request_body = fake_vnflcm.VnfInstances.\
            make_inst_request_body_include_num_dynamic(
                self.vim['tenant_id'], self.ext_networks,
                self.ext_mngd_networks, self.ext_link_ports, self.ext_subnets)
        resp, _ = self._instantiate_vnf_instance(vnf_instance_id, request_body)
        self._wait_lcm_done('COMPLETED', vnf_instance_id=vnf_instance_id)
        self._assert_instantiate_vnf(resp, vnf_instance_id)
        self._assert_stack_template(vnf_instance_id)

        # Show vnf instance
        resp, vnf_instance = self._show_vnf_instance(vnf_instance_id)
        self.assertEqual(200, resp.status_code)

        # Set Fake server response for Grant-Req(Heal)
        vnflcm_base.FAKE_SERVER_MANAGER.set_callback('POST',
            fake_grant.Grant.GRANT_REQ_PATH, status_code=201,
            callback=lambda req_headers,
            req_body: fake_grant.Grant.make_heal_response_body(
                req_body,
                self.vim['tenant_id'],
                glance_image,
                flavour_vdu_dict,
                zone_name_list))

        # Heal vnf (exists vnfc_instace_id)
        vnfc_instance_id_list = []

        for vnfc in vnf_instance.get('instantiatedVnfInfo', {}).\
                get('vnfcResourceInfo', []):
            if vnfc.get('vduId') == 'VDU1':
                vnfc_instance_id_list.append(vnfc.get('id'))

        request_body = fake_vnflcm.VnfInstances.make_heal_request_body(
            vnfc_instance_id_list)
        resp, _ = self._heal_vnf_instance(vnf_instance_id, request_body)
        self._wait_lcm_done('COMPLETED', vnf_instance_id=vnf_instance_id)
        self._assert_heal_vnf(resp, vnf_instance_id)
        self._assert_stack_template(vnf_instance_id)

        # Set Fake server response for Grant-Req(Terminate)
        vnflcm_base.FAKE_SERVER_MANAGER.set_callback('POST',
            fake_grant.Grant.GRANT_REQ_PATH, status_code=201,
            callback=lambda req_headers,
            req_body: fake_grant.Grant.make_term_response_body(req_body))

        # Get stack informations to terminate.
        stack = self._get_heat_stack(vnf_instance_id)
        resources_list = self._get_heat_resource_list(stack.id)
        resource_name_list = [r.resource_name for r in resources_list]
        glance_image_id_list = self._get_glance_image_list_from_stack_resource(
            stack.id, resource_name_list)

        # Terminate VNF
        terminate_req_body = fake_vnflcm.VnfInstances.make_term_request_body()
        resp, _ = self._terminate_vnf_instance(vnf_instance_id,
                terminate_req_body)
        self._wait_lcm_done('COMPLETED', vnf_instance_id=vnf_instance_id)
        self._assert_terminate_vnf(resp, vnf_instance_id, stack.id,
            resource_name_list, glance_image_id_list)

        # Delete VNF
        resp, _ = self._delete_vnf_instance(vnf_instance_id)
        self._wait_lcm_done(vnf_instance_id=vnf_instance_id)
        self.assert_delete_vnf(resp, vnf_instance_id)

        # Delete Subscription
        resp, response_body = self._delete_subscription(subscription_id)
        self.assertEqual(204, resp.status_code)

    def test_inst_scale_term(self):
        """Test basic life cycle operations with sample VNFD with UserData.

        In this test case, we do following steps.
            - Create subscription.
            - Create VNF instance.
            - Instantiate VNF.
            - Get VNF informations.
            - Scale-Out VNF
            - Scale-In VNF
            - Terminate VNF
            - Delete VNF
            - Delete subscription
        """
        vnf_package_info = self._register_vnf_package_mock_response(
            package_dir='functional7')
        glance_image = fake_grant.Grant.get_sw_image(
            package_dir='functional7')
        flavour_vdu_dict = fake_grant.Grant.get_compute_flavor(
            package_dir='functional7')
        zone_name_list = []
        availability_zone_info = self._list_zone()
        zone_name_list = list(set(
            [zone.zone for zone in availability_zone_info
             if zone.binary == 'nova-compute']))

        # Create subscription and register it.
        callback_url = os.path.join(vnflcm_base.MOCK_NOTIFY_CALLBACK_URL,
            self._testMethodName)
        request_body = fake_vnflcm.Subscription.make_create_request_body(
            'http://localhost:{}{}'.format(
                vnflcm_base.FAKE_SERVER_MANAGER.SERVER_PORT,
                callback_url))
        resp, response_body = self._register_subscription(request_body)
        self.assertEqual(201, resp.status_code)
        self.assert_http_header_location_for_subscription(resp.headers)
        self.assert_notification_get(callback_url)
        subscription_id = response_body.get('id')
        self.addCleanup(self._delete_subscription, subscription_id)

        # Create vnf instance
        resp, vnf_instance = self._create_vnf_instance_from_body(
            fake_vnflcm.VnfInstances.make_create_request_body(
                vnf_package_info['vnfdId']))
        vnf_instance_id = vnf_instance.get('id')
        self._wait_lcm_done(vnf_instance_id=vnf_instance_id)
        self._assert_create_vnf(resp, vnf_instance)
        self.addCleanup(self._delete_vnf_instance, vnf_instance_id)

        # Set Fake server response for Grant-Req(Instantiate)
        vnflcm_base.FAKE_SERVER_MANAGER.set_callback('POST',
            fake_grant.Grant.GRANT_REQ_PATH, status_code=201,
            callback=lambda req_headers,
            req_body: fake_grant.Grant.make_inst_response_body(
                req_body,
                self.vim['tenant_id'],
                glance_image,
                flavour_vdu_dict,
                zone_name_list))

        # Instantiate vnf instance
        request_body = fake_vnflcm.VnfInstances.\
            make_inst_request_body_include_num_dynamic(
                self.vim['tenant_id'], self.ext_networks,
                self.ext_mngd_networks, self.ext_link_ports, self.ext_subnets)
        resp, _ = self._instantiate_vnf_instance(vnf_instance_id, request_body)
        self._wait_lcm_done('COMPLETED', vnf_instance_id=vnf_instance_id)
        self._assert_instantiate_vnf(resp, vnf_instance_id)
        self._assert_stack_template_scale(vnf_instance_id)

        # Show vnf instance
        resp, vnf_instance = self._show_vnf_instance(vnf_instance_id)
        self.assertEqual(200, resp.status_code)

        # Set Fake server response for Grant-Req(Scale-out)
        vnflcm_base.FAKE_SERVER_MANAGER.set_callback(
            'POST',
            fake_grant.Grant.GRANT_REQ_PATH, status_code=201,
            callback=lambda req_headers,
            req_body: fake_grant.Grant.make_scale_response_body(
                req_body,
                self.vim['tenant_id'],
                glance_image,
                flavour_vdu_dict,
                zone_name_list))

        # Scale-out vnf instance
        stack = self._get_heat_stack(vnf_instance_id)
        pre_stack_resource_list = self._get_heat_resource_list(stack.id, 2)
        request_body = fake_vnflcm.VnfInstances.\
            make_scale_request_body('SCALE_OUT')
        resp, _ = self._scale_vnf_instance(vnf_instance_id, request_body)
        self._wait_lcm_done('COMPLETED', vnf_instance_id=vnf_instance_id)

        post_stack_resource_list = self._get_heat_resource_list(stack.id, 2)
        self._assert_scale_vnf(resp, vnf_instance_id,
                               pre_stack_resource_list,
                               post_stack_resource_list,
                               scale_type='SCALE_OUT')

        # Set Fake server response for Grant-Req(Scale-in)
        vnflcm_base.FAKE_SERVER_MANAGER.set_callback(
            'POST',
            fake_grant.Grant.GRANT_REQ_PATH, status_code=201,
            callback=lambda req_headers,
            req_body: fake_grant.Grant.make_scale_response_body(
                req_body,
                self.vim['tenant_id'], glance_image, flavour_vdu_dict,
                zone_name_list))

        # Scale-in vnf instance
        stack = self._get_heat_stack(vnf_instance_id)
        pre_stack_resource_list = self._get_heat_resource_list(stack.id, 2)

        request_body = fake_vnflcm.VnfInstances.make_scale_request_body(
            'SCALE_IN')
        resp, _ = self._scale_vnf_instance(vnf_instance_id, request_body)
        self._wait_lcm_done('COMPLETED', vnf_instance_id=vnf_instance_id)

        post_stack_resource_list = self._get_heat_resource_list(stack.id, 2)
        self._assert_scale_vnf(resp, vnf_instance_id,
                               pre_stack_resource_list,
                               post_stack_resource_list,
                               scale_type='SCALE_IN')

        # Set Fake server response for Grant-Req(Terminate)
        vnflcm_base.FAKE_SERVER_MANAGER.set_callback('POST',
            fake_grant.Grant.GRANT_REQ_PATH, status_code=201,
            callback=lambda req_headers,
            req_body: fake_grant.Grant.make_term_response_body(req_body))

        # Get stack informations to terminate.
        stack = self._get_heat_stack(vnf_instance_id)
        resources_list = self._get_heat_resource_list(stack.id)
        resource_name_list = [r.resource_name for r in resources_list]
        glance_image_id_list = self._get_glance_image_list_from_stack_resource(
            stack.id, resource_name_list)

        # Terminate VNF
        terminate_req_body = fake_vnflcm.VnfInstances.make_term_request_body()
        resp, _ = self._terminate_vnf_instance(vnf_instance_id,
                terminate_req_body)
        self._wait_lcm_done('COMPLETED', vnf_instance_id=vnf_instance_id)
        self._assert_terminate_vnf(resp, vnf_instance_id, stack.id,
            resource_name_list, glance_image_id_list)

        # Delete VNF
        resp, _ = self._delete_vnf_instance(vnf_instance_id)
        self._wait_lcm_done(vnf_instance_id=vnf_instance_id)
        self.assert_delete_vnf(resp, vnf_instance_id)

        # Delete Subscription
        resp, response_body = self._delete_subscription(subscription_id)
        self.assertEqual(204, resp.status_code)

    def _assert_create_vnf(self, resp, vnf_instance):
        """Assert that VNF was created via fake server.

        Args:
            resp (Response): HTTP response object.
            vnf_instance (Dict): VNF instance information.
        """
        super().assert_create_vnf(resp, vnf_instance)

        # FT-checkpoint: VnfPkgId
        vnf_pkg_mock_responses = vnflcm_base.FAKE_SERVER_MANAGER.get_history(
            fake_vnfpkgm.VnfPackage.VNF_PACKAGE_REQ_PATH)
        vnflcm_base.FAKE_SERVER_MANAGER.clear_history(
            fake_vnfpkgm.VnfPackage.VNF_PACKAGE_REQ_PATH)

        self.assertEqual(1, len(vnf_pkg_mock_responses))
        vnf_pkg_info_list = vnf_pkg_mock_responses[0]
        self.assertEqual(vnf_instance['vnfPkgId'],
            vnf_pkg_info_list.response_body[0]['id'])

    def _assert_instantiate_vnf(self, resp, vnf_instance_id):
        """Assert that VNF was instantiated.

        This method calls same name method of super class and that
        checks heat resource status 'CREATE_COMPLETE', then assert
        notifications of instantiation.
        Then, we check Grant response in this method.

        Args:
            resp (Response): HTTP response object.
            vnf_instance_id (str): VNF instance id.
        """
        super().assert_instantiate_vnf(resp, vnf_instance_id)

        # FT-checkpoint: Grant Response
        grant_mock_responses = vnflcm_base.FAKE_SERVER_MANAGER.get_history(
            fake_grant.Grant.GRANT_REQ_PATH)
        vnflcm_base.FAKE_SERVER_MANAGER.clear_history(
            fake_grant.Grant.GRANT_REQ_PATH)
        self.assertEqual(1, len(grant_mock_responses))
        self._assert_grant_mock_response(grant_mock_responses[0])

    def _assert_heal_vnf(self, resp, vnf_instance_id,
            expected_stack_status='UPDATE_COMPLETE'):
        """Assert that VNF was healed.

        This method calls same name method of super class and that
        checks heat resource status 'UPDATE_COMPLETE', then assert
        notifications of healing.
        Then, we check Grant response in this method.

        Args:
            resp (Response): HTTP response object.
            vnf_instance_id (str): VNF instance id.
            expected_stack_status (str, optional): self explanatory :)
                                        Defaults to 'UPDATE_COMPLETE'.
        """
        super().assert_heal_vnf(
            resp, vnf_instance_id, expected_stack_status=expected_stack_status)

        # FT-checkpoint: Grant Response
        grant_mock_responses = vnflcm_base.FAKE_SERVER_MANAGER.get_history(
            fake_grant.Grant.GRANT_REQ_PATH)
        vnflcm_base.FAKE_SERVER_MANAGER.clear_history(
            fake_grant.Grant.GRANT_REQ_PATH)
        self.assertEqual(1, len(grant_mock_responses))
        self._assert_grant_mock_response(grant_mock_responses[0])

    def _assert_scale_vnf(self,
            resp,
            vnf_instance_id,
            pre_stack_resource_list,
            post_stack_resource_list,
            scale_type):
        super().assert_scale_vnf(
            resp,
            vnf_instance_id,
            pre_stack_resource_list,
            post_stack_resource_list,
            scale_type=scale_type)

        # FT-checkpoint: Grant Response
        grant_mock_responses = vnflcm_base.FAKE_SERVER_MANAGER.get_history(
            fake_grant.Grant.GRANT_REQ_PATH)
        vnflcm_base.FAKE_SERVER_MANAGER.clear_history(
            fake_grant.Grant.GRANT_REQ_PATH)
        self.assertEqual(1, len(grant_mock_responses))
        self._assert_grant_mock_response(grant_mock_responses[0])

    def _assert_terminate_vnf(self, resp, vnf_instance_id, stack_id,
            resource_name_list, glance_image_id_list):
        """Assert that VNF was terminated.

        This method calls same name method of super class to check specified
        VNF instance is 'NOT_INSTANTIATED'
        Then, we check Grant response in this method.

        Args:
            resp (Response): HTTP response object.
            vnf_instance_id (str): VNF instance id.
            stack_id (str): Resource id of heat stack to check.
            resource_name_list (list[str]): List of heat stack resources.
            glance_image_id_list (list[str]): List of glance image ids.
        """
        super().assert_terminate_vnf(resp, vnf_instance_id, stack_id,
            resource_name_list, glance_image_id_list)

        # FT-checkpoint: Grant Response
        grant_mock_responses = vnflcm_base.FAKE_SERVER_MANAGER.get_history(
            fake_grant.Grant.GRANT_REQ_PATH)
        vnflcm_base.FAKE_SERVER_MANAGER.clear_history(
            fake_grant.Grant.GRANT_REQ_PATH)
        self.assertEqual(1, len(grant_mock_responses))
        self._assert_grant_mock_response(grant_mock_responses[0])

    def _assert_grant_mock_response(self, grant_mock_response,
            expected_auth_type=None, expected_token_value=None):
        """Assert that HTTP response code is equal to 201 or not.

        This method checks response code of grant request and
        authorization result.

        Args:
            grant_mock_response (Response): HTTP response object.
            expected_auth_type (str, optional): Authentication type.
                                                Defaults to None.
            expected_token_value ([type], optional): Authentication token.
                                                     Defaults to None.
        """
        self.assertEqual(201, grant_mock_response.status_code)

        actual_auth = grant_mock_response.request_headers.get("Authorization")
        if expected_auth_type is None:
            self.assertIsNone(actual_auth)
            return

        self.assertEqual(
            '{} {}'.format(expected_auth_type, expected_token_value),
            actual_auth)

    def _assert_occ_show(self, resp, op_occs_info):
        self.assertEqual(200, resp.status_code)

        # Only check required parameters.
        self.assertIsNotNone(op_occs_info.get('id'))
        self.assertIsNotNone(op_occs_info.get('operationState'))
        self.assertIsNotNone(op_occs_info.get('stateEnteredTime'))
        self.assertIsNotNone(op_occs_info.get('vnfInstanceId'))
        self.assertIsNotNone(op_occs_info.get('operation'))
        self.assertIsNotNone(op_occs_info.get('isAutomaticInvocation'))
        self.assertIsNotNone(op_occs_info.get('isCancelPending'))

        _links = op_occs_info.get('_links')
        self.assertIsNotNone(_links.get('self'))
        self.assertIsNotNone(_links.get('self').get('href'))
        self.assertIsNotNone(_links.get('vnfInstance'))
        self.assertIsNotNone(_links.get('vnfInstance').get('href'))
        self.assertIsNotNone(_links.get('grant'))
        self.assertIsNotNone(_links.get('grant').get('href'))

    def _get_fixed_ips(self, vnf_instance_id, request_body):
        res_name = None
        for extvirlink in request_body['extVirtualLinks']:
            if 'extCps' not in extvirlink:
                continue
            for extcps in extvirlink['extCps']:
                if 'cpdId' in extcps:
                    if res_name is None:
                        res_name = list()
                    res_name.append(extcps['cpdId'])
                    break
        self.assertTrue(res_name)

        stack = self._get_heat_stack(vnf_instance_id)
        stack_id = stack.id

        stack_resource = self._get_heat_resource_list(stack_id, nested_depth=2)

        releations = dict()
        for elmt in stack_resource:
            if elmt.resource_type != 'OS::Neutron::Port':
                continue
            if elmt.resource_name not in res_name:
                continue
            releations[elmt.parent_resource] = elmt.resource_name

        details = list()
        for (parent_name, resource_name) in releations.items():
            for elmt in stack_resource:
                if parent_name != elmt.resource_name:
                    continue
                detail_stack = self._get_heat_resource(
                    elmt.physical_resource_id, resource_name)
                details.append(detail_stack)

        ans_list = list()
        for detail in details:
            ans_list.append(detail.attributes['fixed_ips'])

        return ans_list

    def _assert_stack_template(self, vnf_instance_id):
        stack = self._get_heat_stack(vnf_instance_id)
        resources_list\
            = self._get_heat_resource_list(stack.id, nested_depth=2)
        stack_name_wd = vnf_instance_id + "-VDU2"
        physical_resource_id = [r.physical_resource_id for r
        in resources_list if stack_name_wd in r.stack_name]
        template = self._get_heat_stack_template(physical_resource_id[0])
        template_count = str(template).count("flavor")
        self.assertEqual(template_count, 3)

    def _assert_stack_template_scale(self, vnf_instance_id):
        stack = self._get_heat_stack(vnf_instance_id)
        resources_list\
            = self._get_heat_resource_list(stack.id, nested_depth=2)
        stack_name_wd = vnf_instance_id + "-VDU1"
        physical_resource_id = [r.physical_resource_id for r
        in resources_list if stack_name_wd in r.stack_name]
        template = self._get_heat_stack_template(physical_resource_id[0])
        template_count = str(template).count("zone")
        self.assertEqual(template_count, 3)



# Copyright 2016 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""set-description-to-text

Revision ID: c7cde2f45f82
Revises: 6e56d4474b2a
Create Date: 2016-06-01 10:58:43.022668

"""

# flake8: noqa: E402

# revision identifiers, used by Alembic.
revision = 'c7cde2f45f82'
down_revision = '6e56d4474b2a'

from alembic import op
import sqlalchemy as sa


def upgrade(active_plugins=None, options=None):
    op.alter_column('vims',
                    'description',
                    type_=sa.Text)
    op.alter_column('devices',
                    'description',
                    type_=sa.Text)
    op.alter_column('devicetemplates',
                    'description',
                    type_=sa.Text)



# Copyright 2015 Brocade Communications System, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import fixtures

from tacker.common import config
from tacker.db import api as db_api
from tacker.db import model_base
from tacker.tests.unit import base


class SqlFixture(fixtures.Fixture):

    # flag to indicate that the models have been loaded
    _TABLES_ESTABLISHED = False

    def setUp(self):
        super(SqlFixture, self).setUp()
        # Register all data models
        engine = db_api.get_engine()
        if not SqlFixture._TABLES_ESTABLISHED:
            model_base.BASE.metadata.create_all(engine)
            SqlFixture._TABLES_ESTABLISHED = True

        def clear_tables():
            with engine.begin() as conn:
                for table in reversed(
                        model_base.BASE.metadata.sorted_tables):
                    conn.execute(table.delete())

        self.addCleanup(clear_tables)


class SqlTestCase(base.TestCase):

    def setUp(self):
        config.set_db_defaults()
        super(SqlTestCase, self).setUp()
        self.useFixture(SqlFixture())



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tacker.sol_refactored.objects import base
from tacker.sol_refactored.objects import fields


# NFV-SOL 005
# - v3.3.1 9.5.3.2 (API version: 2.1.0)
@base.TackerObjectRegistry.register
class VnfPackageSoftwareImageInfoV2(base.TackerObject,
                                    base.TackerObjectDictCompat):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'id': fields.StringField(nullable=False),
        'name': fields.StringField(nullable=False),
        'provider': fields.StringField(nullable=False),
        'version': fields.VersionField(nullable=False),
        'checksum': fields.ChecksumField(nullable=False),
        'isEncrypted': fields.BooleanField(nullable=False),
        'containerFormat': fields.EnumField(
            valid_values=[
                'AKI',
                'AMI',
                'ARI',
                'BARE',
                'DOCKER',
                'OVA',
                'OVF',
            ],
            nullable=False,
        ),
        'diskFormat': fields.EnumField(
            valid_values=[
                'AKI',
                'AMI',
                'ARI',
                'ISO',
                'QCOW2',
                'RAW',
                'VDI',
                'VHD',
                'VHDX',
                'VMDK',
            ],
            nullable=False,
        ),
        'createdAt': fields.DateTimeField(nullable=False),
        'minDisk': fields.NonNegativeIntegerField(nullable=False),
        'minRam': fields.NonNegativeIntegerField(nullable=False),
        'size': fields.NonNegativeIntegerField(nullable=False),
        'userMetadata': fields.KeyValuePairsField(nullable=True),
        'imagePath': fields.StringField(nullable=True),
        'imageUri': fields.UriField(nullable=True),
    }



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import abc

from keystoneauth1 import adapter
from keystoneauth1 import http_basic
from keystoneauth1.identity import v3
from keystoneauth1 import noauth
from keystoneauth1 import plugin
from keystoneauth1 import session
from oslo_log import log as logging
from oslo_serialization import jsonutils

from tacker.sol_refactored.common import exceptions as sol_ex
from tacker.sol_refactored.objects import base


LOG = logging.getLogger(__name__)


class HttpClient(object):

    def __init__(self, auth_handle, version=None,
            service_type='nfv-orchestration', connect_retries=None,
            timeout=None):
        self.auth_handle = auth_handle
        self.version = version
        self.service_type = service_type
        # NOTE: these parameters could be used from ex. configuration
        # if a HttpClient user want to use these.
        self.connect_retries = connect_retries
        self.timeout = timeout

    def do_request(self, url, method, context=None, expected_status=[],
                   **kwargs):
        content_type = kwargs.pop('content_type', 'application/json')

        headers = kwargs.setdefault('headers', {})
        headers.setdefault('Accept', content_type)

        body = kwargs.pop('body', None)
        if body is not None:
            if isinstance(body, base.TackerObject):
                body = body.to_dict()
            if isinstance(body, dict):
                body = jsonutils.dumps(body)
            kwargs.setdefault('data', body)
            headers.setdefault('Content-Type', content_type)

        version = kwargs.pop('version', None) or self.version
        if version is not None:
            headers.setdefault('Version', version)

        if self.connect_retries is not None:
            kwargs.setdefault('connect_retries', self.connect_retries)
        if self.timeout is not None:
            kwargs.setdefault('timeout', self.timeout)

        session = self.auth_handle.get_session(
            self.auth_handle.get_auth(context), self.service_type)
        resp = session.request(url, method, raise_exc=False, **kwargs)

        resp_body = self._decode_body(resp)

        if expected_status and resp.status_code not in expected_status:
            self.raise_sol_exception(resp, resp_body)

        return resp, resp_body

    def raise_sol_exception(self, resp, resp_body):
        content_type = resp.headers['Content-Type']
        kwargs = {'sol_status': resp.status_code}
        if content_type == 'application/problem+json':
            kwargs['sol_detail'] = resp_body['detail']
        else:
            kwargs['sol_detail'] = resp.text

        raise sol_ex.SolException(**kwargs)

    def _decode_body(self, resp):
        if resp.status_code == 204:  # no content
            return
        content_type = resp.headers['Content-Type']
        if content_type == 'application/zip':
            return resp.content
        if content_type == 'text/plain':
            return resp.text
        if resp.text:
            return jsonutils.loads(resp.text)
        # otherwise return None


class AuthHandle(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def get_auth(self, context=None):
        # returns keystoneauth1 authentication plugin object
        pass

    @abc.abstractmethod
    def get_session(self, auth, service_type):
        # returns keystoneauth1 session object
        pass


class KeystoneTokenAuthHandle(AuthHandle):

    def __init__(self, auth_url, context):
        self.auth_url = auth_url
        self.context = context

    def get_auth(self, context):
        if context is None:
            context = self.context
        return v3.Token(auth_url=self.auth_url,
                        token=context.auth_token,
                        project_id=context.project_id,
                        project_domain_id=context.project_domain_id)

    def get_session(self, auth, service_type):
        _session = session.Session(auth=auth, verify=False)
        return adapter.Adapter(session=_session,
                               service_type=service_type)


class KeystonePasswordAuthHandle(AuthHandle):

    def __init__(self, auth_url, username, password,
            project_name, user_domain_name, project_domain_name):
        self.auth_url = auth_url
        self.username = username
        self.password = password
        self.project_name = project_name
        self.user_domain_name = user_domain_name
        self.project_domain_name = project_domain_name

    def get_auth(self, context=None):
        return v3.Password(auth_url=self.auth_url,
                           username=self.username,
                           password=self.password,
                           project_name=self.project_name,
                           user_domain_name=self.user_domain_name,
                           project_domain_name=self.project_domain_name)

    def get_session(self, auth, service_type):
        _session = session.Session(auth=auth, verify=False)
        return adapter.Adapter(session=_session,
                               service_type=service_type)


class BasicAuthHandle(AuthHandle):

    def __init__(self, username, password):
        self.username = username
        self.password = password

    def get_auth(self, context=None):
        return http_basic.HTTPBasicAuth(username=self.username,
                                        password=self.password)

    def get_session(self, auth, service_type):
        return session.Session(auth=auth, verify=False)


class NoAuthHandle(AuthHandle):

    def __init__(self, endpoint=None):
        self.endpoint = endpoint

    def get_auth(self, context=None):
        return noauth.NoAuth(endpoint=self.endpoint)

    def get_session(self, auth, service_type):
        return session.Session(auth=auth, verify=False)


class Oauth2AuthPlugin(plugin.FixedEndpointPlugin):

    def __init__(self, endpoint, token_endpoint, client_id, client_password):
        super(Oauth2AuthPlugin, self).__init__(endpoint)
        self.token_endpoint = token_endpoint
        self.client_id = client_id
        self.client_password = client_password

    def get_token(self, session, **kwargs):
        auth = BasicAuthHandle(self.client_id,
                               self.client_password)
        client = HttpClient(auth)

        url = self.token_endpoint + '/token'
        data = {'grant_type': 'client_credentials'}

        resp, resp_body = client.do_request(url, "POST",
                data=data, content_type='application/x-www-form-urlencoded')

        if resp.status_code != 200:
            LOG.error("get OAuth2 token failed: %d" % resp.status_code)
            return

        return resp_body['access_token']

    def get_headers(self, session, **kwargs):
        token = self.get_token(session)
        if not token:
            return None
        auth = 'Bearer %s' % token
        return {'Authorization': auth}


class OAuth2AuthHandle(AuthHandle):

    def __init__(self, endpoint, token_endpoint, client_id, client_password):
        self.endpoint = endpoint
        self.token_endpoint = token_endpoint
        self.client_id = client_id
        self.client_password = client_password

    def get_auth(self, context=None):
        return Oauth2AuthPlugin(self.endpoint, self.token_endpoint,
                self.client_id, self.client_password)

    def get_session(self, auth, service_type):
        return session.Session(auth=auth, verify=False)



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from tacker.tests import uuidsentinel


class Subscription:

    @staticmethod
    def make_create_request_body(callback_uri):
        """Parameter selection policy.

        Set all Notification types and all life cycle types for filter.
        Specify OAuth2 for authentication → do not set authentication.

        Args:
            callback_uri (str): Notification URI.

        Returns:
            dict: Request body
        """
        return {
            "filter": {
                "vnfInstanceSubscriptionFilter": {
                    "vnfdIds": ["b1bb0ce7-ebca-4fa7-95ed-4840d7000000"],
                    "vnfProductsFromProviders": [{
                        "vnfProvider": "Company",
                        "vnfProducts": [
                            {
                                "vnfProductName": "Sample VNF",
                                "versions": [
                                    {
                                        "vnfSoftwareVersion": "1.0",
                                        "vnfdVersions": ["1.0"]
                                    }
                                ]
                            }
                        ]
                    }]
                },
                "notificationTypes": [
                    "VnfLcmOperationOccurrenceNotification",
                    "VnfIdentifierCreationNotification",
                    "VnfIdentifierDeletionNotification"
                ],
                "operationTypes": [
                    "INSTANTIATE",
                    "SCALE",
                    "TERMINATE",
                    "HEAL",
                    "MODIFY_INFO",
                    "CHANGE_EXT_CONN"
                ],
                "operationStates": ["STARTING"]
            },
            "callbackUri": callback_uri
        }


ext_vdu1_cp1 = {
    "cpdId": "VDU1_CP1",
    "cpConfig": [{
        "linkPortId": uuidsentinel.elp1_id
    }],
}
ext_vdu2_cp1 = {
    "cpdId": "VDU2_CP1",
    "cpConfig": [{
        "linkPortId": uuidsentinel.elp2_id
    }]
}


def _set_ext_link_port1(external_ports_id):
    ext_link_port1 = {
        "id": uuidsentinel.elp1_id,
        "resourceHandle": {
            "vimConnectionId": uuidsentinel.vim_connection_id,
            "resourceId": external_ports_id[0]
        }
    }
    return ext_link_port1


def _set_ext_link_port2(external_ports_id):
    ext_link_port2 = {
        "id": uuidsentinel.elp2_id,
        "resourceHandle": {
            "vimConnectionId": uuidsentinel.vim_connection_id,
            "resourceId": external_ports_id[1]
        }
    }
    return ext_link_port2


def _set_ext_virtual_link_cp1(networks_id, external_ports_id):
    ext_virtual_link_cp1 = {
        "id": uuidsentinel.evl1_id,
        "resourceId": networks_id[0],
        "vimConnectionId": uuidsentinel.vim_connection_id,
        "extCps": [ext_vdu1_cp1, ext_vdu2_cp1],
        "extLinkPorts": [
            _set_ext_link_port1(external_ports_id),
            _set_ext_link_port2(external_ports_id)]
    }
    return ext_virtual_link_cp1


def _set_ext_cps_vdu1_cp2(external_subnets_id):
    ext_cps_vdu1_cp2 = {
        "cpdId": "VDU1_CP2",
        "cpConfig": [{
            "cpProtocolData": [{
                "layerProtocol": "IP_OVER_ETHERNET",
                "ipOverEthernet": {
                    "ipAddresses": [{
                        "type": "IPV4",
                        "fixedAddresses": ["22.22.1.10"],
                        "subnetId": external_subnets_id[1]
                    }]
                }
            }]
        }]
    }
    return ext_cps_vdu1_cp2


def _set_ext_cps_vdu2_cp2(external_subnets_id):
    ext_cps_vdu2_cp2 = {
        "cpdId": "VDU2_CP2",
        "cpConfig": [{
            "cpProtocolData": [{
                "layerProtocol": "IP_OVER_ETHERNET",
                "ipOverEthernet": {
                    "ipAddresses": [{
                        "type": "IPV4",
                        "fixedAddresses": ["22.22.1.20"],
                        "subnetId": external_subnets_id[1]
                    }]
                }
            }]
        }]
    }
    return ext_cps_vdu2_cp2


def _set_ext_virtual_link_cp2(networks_id, external_subnets_id):
    ext_virtual_link_cp2 = {
        "id": uuidsentinel.evl2_id,
        "resourceId": networks_id[1],
        "vimConnectionId": uuidsentinel.vim_connection_id,
        "extCps": [
            _set_ext_cps_vdu1_cp2(external_subnets_id),
            _set_ext_cps_vdu2_cp2(external_subnets_id)
        ]
    }
    return ext_virtual_link_cp2


def _set_ext_mng_vtl_lnks(ext_mngd_networks_id):
    ext_mng_vtl_lnks = [{
        "id": uuidsentinel.emvl1_id,
        "vnfVirtualLinkDescId": "internalVL1",
        "resourceId": ext_mngd_networks_id[0]
    }, {
        "id": uuidsentinel.emvl2_id,
        "vnfVirtualLinkDescId": "internalVL2",
        "resourceId": ext_mngd_networks_id[1]
    }]
    return ext_mng_vtl_lnks


class VnfInstances:

    @staticmethod
    def make_create_request_body(vnfd_id):
        return {
            "vnfdId": vnfd_id,
            "vnfInstanceName": "",
            "vnfInstanceDescription": "Sample VNF",
            "metadata": {
                "samplekey": "samplevalue"
            }
        }

    @staticmethod
    def make_inst_request_body(
            user_name,
            tenant_id,
            networks_id,
            ext_mngd_networks_id,
            external_ports_id,
            external_subnets_id):
        data = {
            "flavourId": "simple",
            "instantiationLevelId": "instantiation_level_1",
            "extVirtualLinks": [
                _set_ext_virtual_link_cp1(
                    networks_id, external_ports_id),
                _set_ext_virtual_link_cp2(
                    networks_id, external_subnets_id)
            ],
            "extManagedVirtualLinks": _set_ext_mng_vtl_lnks(
                ext_mngd_networks_id),
            "vimConnectionInfo": [{
                "id": uuidsentinel.vim_connection_id,
                "vimType": "ETSINFV.OPENSTACK_KEYSTONE.v_2",
                "vimConnectionId": uuidsentinel.vim_connection_id,
                "interfaceInfo": {
                    "endpoint": "http://127.0.0.1/identity"
                },
                "accessInfo": {
                    "username": user_name,
                    "region": "RegionOne",
                    "password": "devstack",
                    "tenant": tenant_id
                }
            }],
            "additionalParams": {
                "lcm-operation-user-data": "./UserData/lcm_user_data.py",
                "lcm-operation-user-data-class": "SampleUserData"
            }
        }

        return data

    @staticmethod
    def make_term_request_body():
        """Parameter selection policy.

        As all parameters are set, GRACEFUL is specified for terminationType.
        (to specify gracefulTerminationTimeout)

        Returns:
            dict: Request body
        """
        return {
            "terminationType": "GRACEFUL",
            "gracefulTerminationTimeout": 1,
            "additionalParams": {
                "samplekey": "samplevalue"
            }
        }



# Copyright (C) 2019 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import webob
import webob.request

from tacker import context
from tacker.tests import uuidsentinel
from tacker import wsgi as os_wsgi


class FakeRequestContext(context.ContextBaseWithSession):
    def __init__(self, *args, **kwargs):
        kwargs['auth_token'] = kwargs.get(
            uuidsentinel.user_id, uuidsentinel.project_id)
        super(FakeRequestContext, self).__init__(*args, **kwargs)


class HTTPRequest(webob.Request):

    @classmethod
    def blank(cls, *args, **kwargs):
        kwargs['base_url'] = 'http://localhost/'
        use_admin_context = kwargs.pop(context.get_admin_context(), True)
        out = os_wsgi.Request.blank(*args, **kwargs)
        out.environ['tacker.context'] = FakeRequestContext(
            uuidsentinel.user_id,
            uuidsentinel.project_id,
            is_admin=use_admin_context)
        return out



# Copyright (C) 2019 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Internal implementation of request Body validating middleware.

"""

import jsonschema
from jsonschema import exceptions as jsonschema_exc
import netaddr
from oslo_utils import uuidutils
import rfc3986
import webob

from tacker.common import exceptions as exception


@jsonschema.FormatChecker.cls_checks('uri')
def _validate_uri(instance):
    uri = rfc3986.uri_reference(instance)
    validator = rfc3986.validators.Validator().require_presence_of(
        'scheme', 'host',
    ).check_validity_of(
        'scheme', 'userinfo', 'host', 'path', 'query', 'fragment',)
    try:
        validator.validate(uri)
    except rfc3986.exceptions.RFC3986Exception:
        return False
    return True


@jsonschema.FormatChecker.cls_checks('uuid')
def _validate_uuid_format(instance):
    return uuidutils.is_uuid_like(instance)


@jsonschema.FormatChecker.cls_checks('mac_address_or_none',
        webob.exc.HTTPBadRequest)
def validate_mac_address_or_none(instance):
    """Validate instance is a MAC address"""

    if instance is None:
        return

    if not netaddr.valid_mac(instance):
        msg = _("'%s' is not a valid mac address")
        raise webob.exc.HTTPBadRequest(explanation=msg % instance)

    return True


@jsonschema.FormatChecker.cls_checks('mac_address',
        webob.exc.HTTPBadRequest)
def validate_mac_address(instance):
    """Validate instance is a MAC address"""

    if not netaddr.valid_mac(instance):
        msg = _("'%s' is not a valid mac address")
        raise webob.exc.HTTPBadRequest(explanation=msg % instance)

    return True


def _validate_query_parameter_without_value(parameter_name, instance):
    """The query parameter is a flag without a value."""
    if not (isinstance(instance, str) and len(instance)):
        return True

    msg = _("The parameter '%s' is a flag. It shouldn't contain any value.")
    raise webob.exc.HTTPBadRequest(explanation=msg % parameter_name)


@jsonschema.FormatChecker.cls_checks('all_fields',
        webob.exc.HTTPBadRequest)
def _validate_all_fields_query_parameter(instance):
    return _validate_query_parameter_without_value('all_fields', instance)


@jsonschema.FormatChecker.cls_checks('exclude_default',
        webob.exc.HTTPBadRequest)
def _validate_exclude_default_query_parameter(instance):
    return _validate_query_parameter_without_value('exclude_default',
            instance)


class FormatChecker(jsonschema.FormatChecker):
    """A FormatChecker can output the message from cause exception

    We need understandable validation errors messages for users. When a
    custom checker has an exception, the FormatChecker will output a
    readable message provided by the checker.
    """

    def check(self, param_value, format):
        """Check whether the param_value conforms to the given format.

        :argument param_value: the param_value to check
        :type: any primitive type (str, number, bool)
        :argument str format: the format that param_value should conform to
        :raises: :exc:`FormatError` if param_value does not conform to format
        """

        if format not in self.checkers:
            return

        # For safety reasons custom checkers can be registered with
        # allowed exception types. Anything else will fall into the
        # default formatter.
        func, raises = self.checkers[format]
        result, cause = None, None

        try:
            result = func(param_value)
        except raises as e:
            cause = e
        if not result:
            msg = "%r is not a %r" % (param_value, format)
            raise jsonschema_exc.FormatError(msg, cause=cause)


class _SchemaValidator(object):
    """A validator class

    This class is changed from Draft7Validator to validate minimum/maximum
    value of a string number(e.g. '10'). This changes can be removed when
    we tighten up the API definition and the XML conversion.
    Also FormatCheckers are added for checking data formats which would be
    passed through cinder api commonly.

    """
    validator_org = jsonschema.Draft7Validator

    def __init__(self, schema):
        validator_cls = jsonschema.validators.extend(self.validator_org,
                                                     validators={})
        format_checker = FormatChecker()
        self.validator = validator_cls(schema, format_checker=format_checker)

    def validate(self, *args, **kwargs):
        try:
            self.validator.validate(*args, **kwargs)
        except jsonschema.ValidationError as ex:
            if isinstance(ex.cause, webob.exc.HTTPBadRequest):
                detail = str(ex.cause)
            elif len(ex.path) > 0:
                detail = _("Invalid input for field/attribute %(path)s."
                           " Value: %(value)s. %(message)s") % {
                    'path': ex.path.pop(), 'value': ex.instance,
                    'message': ex.message
                }
            else:
                detail = ex.message
            raise exception.ValidationError(detail=detail)
        except TypeError as ex:
            # NOTE: If passing non string value to patternProperties parameter,
            #       TypeError happens. Here is for catching the TypeError.
            raise exception.ValidationError(detail=str(ex))



# All Rights Reserved.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from tacker.api import extensions


class VnfScaleAbstractDriver(extensions.PluginInterface,
        metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def scale(self,
              context,
              plugin,
              auth_attr,
              policy,
              region_name):
        pass

    @abc.abstractmethod
    def scale_wait(self,
                   context,
                   plugin,
                   auth_attr,
                   policy,
                   region_name):
        pass

    @abc.abstractmethod
    def get_scale_ids(self,
                      plugin,
                      context,
                      vnf_dict,
                      auth_attr,
                      region_name):
        pass

    @abc.abstractmethod
    def get_scale_in_ids(self,
                         plugin,
                         context,
                         vnf_dict,
                         is_reverse,
                         auth_attr,
                         region_name,
                         number_of_steps):
        pass

    @abc.abstractmethod
    def scale_resource_update(self, context, vnf_instance,
                              scale_vnf_request, vnf_info,
                              vim_connection_info):
        pass

    @abc.abstractmethod
    def scale_in_reverse(self,
              context,
              plugin,
              auth_attr,
              vnf_info,
              scale_vnf_request,
              region_name,
              scale_name_list,
              grp_id,
              vnf_instance):
        pass

    @abc.abstractmethod
    def scale_out_initial(self,
              context,
              plugin,
              auth_attr,
              vnf_info,
              scale_vnf_request,
              region_name,
              vnf_instance):
        pass

    @abc.abstractmethod
    def scale_update_wait(self,
                   context,
                   plugin,
                   auth_attr,
                   vnf_info,
                   region_name):
        pass

    @abc.abstractmethod
    def get_cinder_list(self,
                        vnf_info):
        pass

    @abc.abstractmethod
    def get_grant_resource(self,
                   plugin,
                   vnf_instance,
                   vnf_info,
                   scale_vnf_request,
                   placement_obj_list,
                   vim_connection_info,
                   del_list):
        pass

    @abc.abstractmethod
    def get_rollback_ids(self,
                         plugin,
                         context,
                         vnf_dict,
                         aspect_id,
                         auth_attr,
                         region_name):
        pass



# Copyright (C) 2019 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Common parameter types for validating request Body.

"""

import re
import unicodedata


def _is_printable(char):
    """determine if a unicode code point is printable.

    This checks if the character is either "other" (mostly control
    codes), or a non-horizontal space. All characters that don't match
    those criteria are considered printable; that is: letters;
    combining marks; numbers; punctuation; symbols; (horizontal) space
    separators.
    """
    category = unicodedata.category(char)
    return (not category.startswith("C") and
            (not category.startswith("Z") or category == "Zs"))


def _get_all_chars():
    for i in range(0xFFFF):
        yield chr(i)


# build a regex that matches all printable characters. This allows
# spaces in the middle of the name. Also note that the regexp below
# deliberately allows the empty string. This is so only the constraint
# which enforces a minimum length for the name is triggered when an
# empty string is tested. Otherwise it is not deterministic which
# constraint fails and this causes issues for some unittests when
# PYTHONHASHSEED is set randomly.

def _build_regex_range(ws=True, invert=False, exclude=None):
    """Build a range regex for a set of characters in utf8.

    This builds a valid range regex for characters in utf8 by
    iterating the entire space and building up a set of x-y ranges for
    all the characters we find which are valid.

    :param ws: should we include whitespace in this range.
    :param invert: invert the logic
    :param exclude: any characters we want to exclude

    The inversion is useful when we want to generate a set of ranges
    which is everything that's not a certain class. For instance,
    produce all all the non printable characters as a set of ranges.
    """
    if exclude is None:
        exclude = []
    regex = ""
    # are we currently in a range
    in_range = False
    # last character we found, for closing ranges
    last = None
    # last character we added to the regex, this lets us know that we
    # already have B in the range, which means we don't need to close
    # it out with B-B. While the later seems to work, it's kind of bad form.
    last_added = None

    def valid_char(char):
        if char in exclude:
            result = False
        elif ws:
            result = _is_printable(char)
        else:
            # Zs is the unicode class for space characters, of which
            # there are about 10 in this range.
            result = (_is_printable(char) and
                      unicodedata.category(char) != "Zs")
        if invert is True:
            return not result
        return result

    # iterate through the entire character range. in_
    for c in _get_all_chars():
        if valid_char(c):
            if not in_range:
                regex += re.escape(c)
                last_added = c
            in_range = True
        else:
            if in_range and last != last_added:
                regex += "-" + re.escape(last)
            in_range = False
        last = c
    else:
        if in_range:
            regex += "-" + re.escape(c)
    return regex


valid_description_regex_base = '^[%s]*$'

valid_description_regex = valid_description_regex_base % (
    _build_regex_range())


keyvalue_pairs = {
    'type': 'object',
    'patternProperties': {
        '^[a-zA-Z0-9-_:. /]{1,255}$': {
            'anyOf': [
                {'type': 'array'},
                {'type': 'string', 'maxLength': 255},
                {'type': 'object'},
                {'type': 'null'},
                {'type': 'boolean'}
            ]
        }
    },
    'additionalProperties': False
}

description = {
    'type': 'string', 'minLength': 0, 'maxLength': 1024,
    'pattern': valid_description_regex,
}

uuid = {
    'type': 'string', 'format': 'uuid'
}

name_allow_zero_min_length = {
    'type': 'string', 'minLength': 0, 'maxLength': 255
}

ip_address = {
    'type': 'string',
    'oneOf': [
        {'format': 'ipv4'},
        {'format': 'ipv6'}
    ]
}

positive_integer = {
    'type': ['integer', 'string'],
    'pattern': '^[0-9]*$', 'minimum': 1, 'minLength': 1
}

identifier = {
    'type': 'string', 'minLength': 1, 'maxLength': 255
}

identifier_in_vim = {
    'type': 'string', 'minLength': 1, 'maxLength': 255
}

identifier_in_vnfd = {
    'type': 'string', 'minLength': 1, 'maxLength': 255
}

identifier_in_vnf = {
    'type': 'string', 'minLength': 1, 'maxLength': 255
}

mac_address_or_none = {
    'type': 'string', 'format': 'mac_address_or_none'
}
